"""
This subpackage contains one python source file per format implemented for parsing.
"""

# Built-in modules #
import os, sys

# Variables #
parsers = {
    'memory':   {'module': 'track.parse.memory',   'class': 'ParserMemory'},
    'sql':      {'module': 'track.parse.sql',      'class': 'ParserSQL'},
    'bed':      {'module': 'track.parse.bed',      'class': 'ParserBED'},
    'wig':      {'module': 'track.parse.wig',      'class': 'ParserWIG'},
    'gff':      {'module': 'track.parse.gff',      'class': 'ParserGFF'},
    'gtf':      {'module': 'track.parse.gtf',      'class': 'ParserGTF'},
    'bedgraph': {'module': 'track.parse.bedgraph', 'class': 'ParserBedgraph'},
    'bigwig':   {'module': 'track.parse.bigwig',   'class': 'ParserBigwig'},
    'sga':      {'module': 'track.parse.sga',      'class': 'ParserSGA'},
}

################################################################################
def get_parser(path, format):
    """Given a path and a format will return the appropriate parser.

            * *path* is a string specifying the path of the track to parse.
            * *format* is a string specifying the format of the track to parse.

        Examples::

            import track.parse
            import track.serialze
            serializer = track.serialize.get_serializer('tmp/test.sql', 'sql')
            parser = track.parse.get_parser('tmp/test.bed', 'bed')
            parser(serializer)

        ``get_parser`` returns a Parser instance.
    """
    # Is the format supported ? #
    if not format in parsers: raise Exception("The format '%s' is not supported." % format)
    info = parsers[format]
    # Import the objects #
    base_module    = __import__(info['module'])
    sub_module     = sys.modules[info['module']]
    class_object   = getattr(sub_module, info['class'])
    class_instance = class_object(path)
    # Return an instance #
    return class_instance

################################################################################
class Parser(object):
    def __init__(self, path):
        self.path = path
        if isinstance(path, basestring): self.name = os.path.basename(path)
        else: self.name = path.name

    def __call__(self, handler=None):
        # Default handler #
        if not handler:
            from track.seralize.memory import SerializerMemory
            handler = SerializerMemory()
        # Enter the handler #
        with handler as self.handler: self.parse()
        # Return a list of paths or a single path #
        if len(self.handler.tracks) == 1: return self.handler.tracks[0]
        else: return self.handler.tracks

    def parse(self):
        raise NotImplementedError

#-----------------------------------#
# This code was written by the BBCF #
# http://bbcf.epfl.ch/              #
# webmaster.bbcf@epfl.ch            #
#-----------------------------------#



"""
Extra functionality that can be used by other libraries.
"""

# Internal modules #
import track
from track.util import add_chromsome_prefix
from track.common import natural_sort, collapse

################################################################################
class TrackCollection(object):
    """
    The TrackCollection class can be useful for grouping many
    Track objects together and represent them as one.
    """
    def __init__(self, tracks, fields_collapse=None, chroms_collapse=None):
        self.tracks = tracks
        self.fields_collapse = fields_collapse
        self.chroms_collapse = chroms_collapse

    def __iter__(self): return iter(self.chromosomes)
    def __contains__(self, key): return key in self.chromosomes
    def __len__(self): return len(self.chromosomes)
    def __nonzero__(self): return True
    def __repr__(self): return "TrackCollection containing %s" % self.tracks

    @property
    def chromosomes(self):
        if not hasattr(self.tracks[0], "chromosomes"): return []
        chroms = collapse(self.chroms_collapse, [t.chromosomes for t in self.tracks])
        chroms.sort(key=natural_sort)
        return chroms

    @property
    def name(self):
        plural = len(self.tracks) > 1
        return 'Collection of %i track%s' % (len(self.tracks), plural and 's' or '')

    @property
    def chrmeta(self):
        if not hasattr(self.tracks[0], "chrmeta"): return {}
        return dict([(chrom,meta) for chrom,meta in self.tracks[0].chrmeta if chrom in self])

    @property
    def datatype(self):
        for t in self.tracks:
            if t.datatype: return t.datatype

    @property
    def assembly(self):
        for t in self.tracks:
            if t.assembly: return t.assembly

    @property
    def fields(self):
        return collapse(self.fields_collapse, [t.fields for t in self.tracks])

    @fields.setter
    def fields(self, value):
        for t in self.tracks: t.fields = value

    def read(self, *args, **kwargs):
        return [t.read(*args, **kwargs) for t in self.tracks]

    def close(self):
        for t in self.tracks: t.close()

################################################################################
class VirtualTrack(object):
    """
    The VirtualTrack class is a promise of a track to come in the future. The VirtualTrack
    has methods similar to the real Track. You can query its *fields* attribute and read
    specific chromosomes. The thing is that the features don't exist until you try to read them.
    Only once you issue ``virtual_track.read('chr1')`` is the result computed just
    in time.
    """
    def __init__(self):
        self.promises = {}
        self.chrmeta = {}
        self.info = {}
        self.tracks_to_close = []

    def __iter__(self): return iter(self.chromosomes)
    def __contains__(self, key): return key in self.chromosomes
    def __len__(self): return len(self.chromosomes)
    def __nonzero__(self): return True

    @property
    def chromosomes(self):
        chroms = self.promises.keys()
        chroms.sort(key=natural_sort)
        return chroms

    @property
    def datatype(self):
        return self.info.get('datatype', None)

    @datatype.setter
    def datatype(self, value):
        self.info['datatype'] = value

    @property
    def name(self):
        return self.info.get('name', 'Unnamed')

    @name.setter
    def name(self, value):
        self.info['name'] = value

    @property
    def assembly(self):
        return self.info.get('assembly', None)

    @assembly.setter
    def assembly(self, value):
        self.info['assembly'] = value

    def write(self, chromosome, stream):
        self.promises[chromosome] = stream

    def read(self, chromosome=None):
        # If we have a specific chromosome #
        if chromosome: return self.promises[chromosome]
        # Else we want all the chromosomes at once #
        return self.read_all()

    def read_all(self):
        for chrom in self:
            for f in add_chromsome_prefix(chrom, self.read(chrom)): yield f

    def export(self, path, format=None):
        with track.new(path, format) as t:
            for chrom in self: t.write(chrom, self.read(chrom))
            t.chrmeta = self.chrmeta
            t.info    = self.info
        self.close()

    def close(self):
        for t in self.tracks_to_close: t.close()



"""
This module implements the GFF serialization.
"""

# Internal modules #
from track.serialize import Serializer
from track.common import format_float
from track.util import int_to_strand

# Constants #
all_fields = ['source', 'name', 'start', 'end', 'score', 'strand', 'frame', 'attributes']
defaults   = ['.', '.', -1, -1, 0.0, 0, '.', '.']

################################################################################
class SerializerGFF(Serializer):
    format = 'gff'

    def __enter__(self):
        self.file = open(self.path, 'w')
        self.indices = []
        return self

    def __exit__(self, errtype, value, traceback):
        self.file.close()

    def defineFields(self, fields):
        for f in all_fields:
            try: self.indices.append(fields.index(f))
            except ValueError: self.indices.append(-1)

    def newTrack(self, info=None, name=None):
        if not info: info = {}
        info['type'] = 'gff'
        info['converted_by'] = __package__
        self.file.write("track " + ' '.join([k + '="' + v + '"' for k, v in info.items()]) + '\n')
        self.tracks.append(self.path)

    def newFeature(self, chrom, feature):
        # Put the fields in the right order #
        line = range(len(all_fields))
        for n,i in enumerate(self.indices):
            if i == -1: line[n] = defaults[n]
            else:       line[n] = feature[i]
        # Convert the score #
        line[4] = format_float(line[4])
        # Convert the strand #
        line[5] = int_to_strand(line[5])
        # Convert the frame #
        if line[6] == '' or line[6] == None: line[6] = '.'
        # Make sure eveything is a string #
        line = [str(f) for f in line]
        # Write one line #
        self.file.write('\t'.join([chrom] + line) + '\n')

#-----------------------------------#
# This code was written by the BBCF #
# http://bbcf.epfl.ch/              #
# webmaster.bbcf@epfl.ch            #
#-----------------------------------#



"""
This module implements the BED serialization.
"""

# Internal modules #
from track.serialize import Serializer
from track.common import format_float
from track.util import int_to_strand

# Constants #
all_fields = ['start', 'end', 'name', 'score', 'strand', 'thick_start',
              'thick_end', 'item_rgb', 'block_count', 'block_sizes', 'block_starts']

################################################################################
class SerializerBED(Serializer):
    format = 'bed'

    def __enter__(self):
        self.file = open(self.path, 'w')
        self.indices = []
        return self

    def __exit__(self, errtype, value, traceback):
        self.file.close()

    def defineFields(self, fields):
        # Check that all fields are legal #
        if not all([f in all_fields for f in fields]):
            message = "You tried to write a '%s' fields to a BED file. Possible fields are: %s"
            self.error(message % (fields, all_fields))
        # Check for the simple case  #
        if all([fields[i] == all_fields[i] for i in range(len(fields))]): return
        # Find the highest ranked field #
        number_of_columns = max(all_fields.index(f) for f in fields) + 1
        # Check for availability #
        for f in all_fields[:number_of_columns]:
            if f in fields: self.indices.append(fields.index(f))
            else: self.indices.append(None)

    def newTrack(self, info=None, name=None):
        if not info: info = {}
        info['type'] = 'bed'
        info['converted_by'] = __package__
        self.file.write("track " + ' '.join([k + '="' + v + '"' for k, v in info.items()]) + '\n')
        self.tracks.append(self.path)

    def newFeature(self, chrom, feature):
        # Put the fields in the right order #
        if self.indices: line = [feature[i] if i != None else '' for i in self.indices]
        else:            line = list(feature)
        # Convert the score #
        try: line[3] = format_float(line[3])
        except IndexError: pass
        # Convert the strand #
        try: line[4] = int_to_strand(line[4])
        except IndexError: pass
        # Make sure eveything is a string #
        line = [str(f) for f in line]
        # Write one line #
        self.file.write('\t'.join([chrom] + line) + '\n')

#-----------------------------------#
# This code was written by the BBCF #
# http://bbcf.epfl.ch/              #
# webmaster.bbcf@epfl.ch            #
#-----------------------------------#



"""
Contains tests for the track.search method.
"""

# Internal modules #
import track
from track.test import samples

# Unittesting module #
try:
    import unittest2 as unittest
except ImportError:
    import unittest

# Nosetest flag #
__test__ = True

###################################################################################
class TestVector(unittest.TestCase):
    """A simple case without boundaries"""
    def runTest(self):
        in_path = samples['small_features'][4]['sql']
        with track.load(in_path) as t:
            data = t.get_full_score_vector('chrI')
            got = list(data)
        expected = [0.0]*10 + [1.0]*10 + [0.0]*10 + [2.0]*10 + [0.0]*230168
        self.assertEqual(got, expected)

###################################################################################
class TestNoEnd(unittest.TestCase):
    """A simple case on a track without specifying an end"""
    def runTest(self):
        in_path = samples['small_features'][4]['sql']
        with track.load(in_path, readonly=True) as t:
            t.chrmeta = {}
            data = t.get_full_score_vector('chrI')
            got = list(data)
        expected = [0.0]*10 + [1.0]*10 + [0.0]*10 + [2.0]*10
        self.assertEqual(got, expected)

###################################################################################
class TestBoundaries(unittest.TestCase):
    """A simple case with boundaries"""
    def runTest(self):
        in_path = samples['small_features'][4]['sql']
        with track.load(in_path) as t:
            data = t.get_partial_score_vector('chrI', 15, 35)
            got = list(data)
        expected = [1.0]*5 + [0.0]*10 + [2.0]*5
        self.assertEqual(got, expected)

###################################################################################
class TestNoScore(unittest.TestCase):
    """A simple case on a track without scores"""
    def runTest(self):
        in_path = samples['small_features'][5]['sql']
        with track.load(in_path) as t:
            data = t.get_partial_score_vector('chrI', 10, 30)
            got = list(data)
        expected = [0.0]*4 + [1.0]*5 + [0.0]*8 + [1.0]*3
        self.assertEqual(got, expected)

###################################################################################
class TestOverlaps(unittest.TestCase):
    """A case on a feature track that has internal overlaps"""
    def runTest(self):
        in_path = samples['small_features'][1]['sql']
        with track.load(in_path) as t:
            data = t.get_partial_score_vector('chrI', 0, 50)
            got = list(data)
        expected = [10.0]*10 + [0.0]*10 + [10.0]*10 + [0.0]*15 + [10.0]*5
        self.assertEqual(got, expected)

#-----------------------------------#
# This code was written by the BBCF #
# http://bbcf.epfl.ch/              #
# webmaster.bbcf@epfl.ch            #
#-----------------------------------#



# coding: utf-8

################################### Modules ####################################
import sys
from track import common
from track.test import samples

################################# Description ##################################
label           = 'filter'
short_name      = 'Filter'
long_name       = 'Filter features in a track by using a second track'
input_tracks    = [{'key':'X', 'position':1, 'kind':'single', 'fields':['start','end','...']},
                   {'key':'Y', 'position':2, 'kind':'single', 'fields':['start','end']}]
input_args      = []
input_meta      = []
output_tracks   = [{'position':1, 'fields': ['start','end','...']}]
tracks_collapse = None
fields_collapse = None
chroms_collapse = 'first'

################################ Documentation #################################
tooltip = \
"""
The ``filter`` manipulation computes the overlap of
the first track against the second stream returning only
complete features from the first track.
"""

tickets = []

################################## Examples ####################################
numeric_example = \
"""
X [start,end]: (10,20) (30,40)
Y [start,end]: (10,12) (17,22)
R [start,end]: (10,20) (30,40)
"""

visual_example = \
u"""
X: ───▤▤▤▤▤▤▤▤▤────────────▤▤▤▤▤▤▤──────
Y: ──────▤▤▤▤▤────▤▤▤▤▤─────────────────
R: ───▤▤▤▤▤▤▤▤▤─────────────────────────
"""

#################################### Tests #####################################
tests = [
    {'tracks':     {'X': samples['small_features'][2]['sql'],
                    'Y': samples['small_features'][3]['sql']},
     'chromosome': 'chrI',
     'expected':  [( 10,  20, u'Name1',  0.1, -1),
                   ( 30,  40, u'Name2',  0.2, 1),
                   ( 90, 100, u'Name5',  0.0, 1),
                   (110, 120, u'Name6',  0.4, 1),
                   (130, 150, u'Name7',  0.4, -1),
                   (180, 190, u'Name8',  0.1, 1),
                   (180, 200, u'Name9',  0.1, 1),
                   (210, 220, u'Name10', 0.2, 1),
                   (230, 240, u'Name11', 0.1, 1),
                   (250, 260, u'Name12', 0.2, 1),
                   (270, 280, u'Name13', 0.0, 1)]}]

############################### Implementation #################################
def generate(X, Y):
    sentinel = (sys.maxint, sys.maxint)
    X = common.sentinelize(X, sentinel)
    Y = common.sentinelize(Y, sentinel)
    x = X.next()
    y = Y.next()
    if x == sentinel or y == sentinel: continue_loop = False
    else:                              continue_loop = True
    while continue_loop:
        open_window  = y[0]
        close_window = y[1]
        # Extend the Y window as long as possible #
        while True:
            if y == sentinel: break
            y = Y.next()
            if y[0] > close_window: break
            if y[1] > close_window: close_window = y[1]
        # Read features from X until overshooting the Y window #
        while True:
            if x[0] >= close_window: break
            if x[1] >  open_window:  yield x
            x = X.next()
            if x == sentinel:
                continue_loop = False
                break



"""
Contains tests for the SuperRow object.
"""

# Built-in modules #
import sqlite3

# Internal modules #
from track.pyrow import SuperRow

# Unittesting module #
try:
    import unittest2 as unittest
except ImportError:
    import unittest

# Nosetest flag #
__test__ = True

###################################################################################
class TestRowSlice(unittest.TestCase):
    def runTest(self):
        con = sqlite3.connect(":memory:")
        con.row_factory = SuperRow
        row = con.execute("select 1, 2, 3, 4").fetchone()
        self.assertTrue(row[0:0]   == (),      "by slice: wrong result for [0:0]")
        self.assertTrue(row[0:1]   == (1,),    "by slice: wrong result for [0:1]")
        self.assertTrue(row[1:3]   == (2,3),   "by slice: wrong result for [1:3]")
        self.assertTrue(row[1:]    == (2,3,4), "by slice: wrong result for [1:]")
        self.assertTrue(row[:3]    == (1,2,3), "by slice: wrong result for [:3]")
        self.assertTrue(row[-2:-1] == (3,),    "by slice: wrong result for [-2:-1]")
        self.assertTrue(row[-2:]   == (3,4),   "by slice: wrong result for [-2:]")

#-----------------------------------#
# This code was written by the BBCF #
# http://bbcf.epfl.ch/              #
# webmaster.bbcf@epfl.ch            #
#-----------------------------------#



# coding: utf-8

################################### Modules ####################################
import sys
from track import common

################################# Description ##################################
label           = 'window_smoothing'
short_name      = 'Window smoothing'
long_name       = 'Smooth scores with a moving window'
input_tracks    = [{'key':'X', 'position':1, 'kind':'single', 'no-overlap':True,
                    'fields':['start','end','score']}]
input_args      = [{'key':'L', 'position':2, 'type': int, 'default': 200, 'doc':'The window radius.'}]
input_meta      = [{'key':'l', 'position':3, 'type': int, 'kind':'chrom_len'}]
output_tracks   = [{'position':1, 'fields': ['start','end','score'], 'datatype':'signal'}]
tracks_collapse = None
fields_collapse = None
chroms_collapse = None

################################ Documentation #################################
tooltip = \
"""
Given a signal track and a window size in base pairs,
the ``windows_smoothing`` manipulation will output a new signal
track with, at each position p, the mean of the scores
in the window [p-L, p+L]. Border cases are handled by zero padding
and the signal's support is invariant.
"""

tickets = ['https://github.com/bbcf/gMiner/issues/7']

################################## Examples ####################################
numeric_example = \
""""""

visual_example = \
u"""
X: ▁▁▁▁▁▁▁▁▁▁████████████▁▁▁▁▁▁▁▁▁▁▁▁
R: ▁▁▁▁▁▁▂▄▅▇████████████▇▅▄▂▁▁▁▁▁▁▁▁
"""

#################################### Tests #####################################
tests = [
    {'tracks':   {'X': iter([(0,2,10),(2,4,20),(6,8,10)])},
     'args':     {'l':9, 'L':2},
     'expected': [(0,  1,   8.0),
                  (1,  3,  12.0),
                  (3,  5,  10.0),
                  (5,  6,   8.0),
                  (6,  9,   4.0),]}
     ,
    {'tracks':   {'X': iter([(0,2,10),(2,4,20),(6,8,10)])},
     'args':     {'l':12, 'L':2},
     'expected': [(0,  1,   8.0),
                  (1,  3,  12.0),
                  (3,  5,  10.0),
                  (5,  6,   8.0),
                  (6,  9,   4.0),
                  (9,  10,  2.0),]}]

############################### Implementation #################################
def generate(X, L, l):
    # Sentinel #
    sentinel = (sys.maxint, sys.maxint, 0.0)
    X = common.sentinelize(X, sentinel)
    # Growing and shrinking list of features around our moving window #
    F = []
    # Current position counting on nucleotides (first nucleotide is zero) #
    p = -L-2
    # Position since which the mean hasn't changed #
    same_since = -L-3
    # The current mean and the next mean #
    curt_mean = 0
    next_mean = 0
    # Multiplication factor instead of division #
    f = 1.0 / (2*L+1)
    # First feature if it exists #
    F.append(X.next())
    if F == [sentinel]: return
    # Core loop #
    while True:
        # Advance one #
        p += 1
        # Window start and stop #
        s = p-L
        e = p+L+1
        # Scores entering window #
        if F[-1][1] < e: F.append(X.next())
        if F[-1][0] < e: next_mean += F[-1][2] * f
        # Scores exiting window #
        if F[ 0][1] < s: F.pop(0)
        if F[ 0][0] < s: next_mean -= F[ 0][2] * f
        # Border condition on the left #
        if p < 0:
            curt_mean = 0
            same_since = p
            continue
        # Border condition on the right #
        if p == l:
            if curt_mean != 0: yield (same_since, p, curt_mean)
            break
        # Maybe emit a feature #
        if next_mean != curt_mean:
            if curt_mean != 0: yield (same_since, p, curt_mean)
            curt_mean  = next_mean
            same_since = p



# coding: utf-8

################################### Modules ####################################
from track.test import samples

################################# Description ##################################
label           = 'closest_features'
short_name      = 'Closest Features'
long_name       = 'Find the closest features from one track in an other track'
input_tracks    = [{'key':'X', 'position':1, 'kind':'single',
                    'fields':['start','end','name']},
                   {'key':'Y', 'position':2, 'kind':'single',
                    'fields':['start','end','name','strand']}]
input_args      = [{'key':'min_length',  'position':3, 'type': int, 'default': 100000,
                    'doc':'Minimal distance for attribution.'},
                   {'key':'utr_cutoff',  'position':4, 'type': int, 'default': 2000,
                    'doc':'Distance below which the gene is attributed to both.'},
                   {'key':'prom_cutoff', 'position':5, 'type': int, 'default': 10,
                    'doc':'Distance percentage below which attribution is to preceding gene.'}]
input_meta      = []
output_tracks   = [{'position':1, 'fields': ['start','end','name','id','type','location']}]
tracks_collapse = None
fields_collapse = None
chroms_collapse = 'first'

################################ Documentation #################################
tooltip = \
"""
For instance, the ``closest_features`` manipulation can identify
the most relevant(s) gene(s) associated to each peak.

* In cases where the peak is isolated, ``min_length`` indicates maximal allowed length between peak and closest gene. Beyond this distance, the peak will not be attributed to any gene. This defaults to ``100000``.

* In cases where the peak is near two promoters (one on each strand), ``utr_cutoff`` indicates the threshold above which the peak isn't attributed to any gene. Below this threshold, the peak will be attributed to both genes. This defaults to ``2000``.

* In cases where the peak is between two genes on the same strand, ``prom_cutoff`` indicates the percentage of the distance between the two genes below which the peak is attributed to the 3'UTR of the preceding gene rather than to the promoter of the following gene. This defaults to ``10``.
"""

tickets = []

################################## Examples ####################################
numeric_example = \
""""""

visual_example = \
u"""
                             peak
      ______  min_length    ------       min_length  ______
-----|______|------...------------------...---------|______|-----
      gene 1                                         gene 2

          -->                          -->
         |______  10%        90%      |______
---------|______|-----|---------------|______|-------------------
          gene 1      prom_cutoff      gene 2
"""

#################################### Tests #####################################
tests = [
    {'tracks':   {'X': samples['small_signals'][4]['sql'],
                  'Y': samples['small_features'][2]['sql']},
     'chromosome': 'chrI',
     'expected': [(10, 20, 'Name1',  15.0, -1),
                  (30, 40, 'Name2',  50.0,  1)]}]

############################### Implementation #################################
def generate(X, Y, min_length, utr_cutoff, prom_cutoff):
    yield (10, 20, 'Name1',  15.0, -1)
    yield (30, 40, 'Name2',  50.0,  1)



"""
This subpackage contains one python source file per format implemented for serialization.
"""

# Built-in modules #
import sys

# Other modules #
from track.genomes import Assembly

# Variables #
serializers = {
    'memory':   {'module': 'track.serialize.memory',   'class': 'SerializerMemory'},
    'sql':      {'module': 'track.serialize.sql',      'class': 'SerializerSQL'},
    'bed':      {'module': 'track.serialize.bed',      'class': 'SerializerBED'},
    'wig':      {'module': 'track.serialize.wig',      'class': 'SerializerWIG'},
    'gff':      {'module': 'track.serialize.gff',      'class': 'SerializerGFF'},
    'gtf':      {'module': 'track.serialize.gtf',      'class': 'SerializerGTF'},
    'bedgraph': {'module': 'track.serialize.bedgraph', 'class': 'SerializerBedgraph'},
    'bigwig':   {'module': 'track.serialize.bigwig',   'class': 'SerializerBigwig'},
    'sga':      {'module': 'track.serialize.sga',      'class': 'SerializerSGA'},
}

################################################################################
def get_serializer(path, format):
    """Given a path and a format will return the appropriate serializer.

            * *path* is a string specifying the path of the track to parse.
            * *format* is a string specifying the format of the track to parse.

        Examples::

            import track.parse
            import track.serialze
            serializer = track.serialize.get_serializer('tmp/test.sql', 'sql')
            parser = track.parse.get_parser('tmp/test.bed', 'bed')
            parser(serializer)

        ``get_serializer`` returns a Serializer instance.
    """
    # Is the format supported ? #
    if not format in serializers: raise Exception("The format '%s' is not supported." % format)
    info = serializers[format]
    # Import the objects #
    base_module    = __import__(info['module'])
    sub_module     = sys.modules[info['module']]
    class_object   = getattr(sub_module, info['class'])
    class_instance = class_object(path)
    # Return an instance #
    return class_instance

################################################################################
class Serializer(object):
    def __init__(self, path):
        self.path = path
        self.tracks = []

    def __enter__(self):
        return self

    def __exit__(self, errtype, value, traceback):
        pass

    def __call__(self, parser):
        self.parser = parser

    def error(self, message, path=None, line_number=None):
        if path and line_number:       message % (" '%s:%s'" % (path, line_number))
        elif path and not line_number: message % (" '%s'"    % (path))
        raise Exception(message)

    def defineFields(self, fields):
        pass

    def defineChrmeta(self, chrmeta):
        self.chrmeta = chrmeta

    def defineAssembly(self, assembly):
        self.assembly = Assembly(assembly)
        self.defineChrmeta(self.assembly.chrmeta)

    def newTrack(self, info=None, name=None):
        pass

    def newFeature(self, chrom, feature):
        raise NotImplementedError

#-----------------------------------#
# This code was written by the BBCF #
# http://bbcf.epfl.ch/              #
# webmaster.bbcf@epfl.ch            #
#-----------------------------------#



"""
Contains tests for the BED format.
We will do a roundtrips for several sample validation tracks.
"""

# Built-in modules #
import os

# Internal modules #
import track
from track.common import temporary_path, assert_file_equal
from track.test import samples

# Unittesting module #
try:
    import unittest2 as unittest
except ImportError:
    import unittest

# Nosetest flag #
__test__ = True

###################################################################################
class TestConversion(unittest.TestCase):
    def runTest(self):
        for num, info in sorted(samples['yeast_features'].items()):
            # Prepare paths #
            orig_bed_path = info['bed']
            orig_sql_path = info['sql']
            test_sql_path = temporary_path('.sql')
            # From BED to SQL #
            track.convert(orig_bed_path, test_sql_path, assembly='sacCer2')
            self.assertTrue(assert_file_equal(orig_sql_path, test_sql_path))
            # Clean up #
            os.remove(test_sql_path)

class TestRoundtrip(unittest.TestCase):
    def runTest(self):
        for num, info in sorted(samples['small_features'].items()):
            # Prepare paths #
            orig_bed_path = info['bed']
            orig_sql_path = info['sql']
            test_sql_path = temporary_path('.sql')
            test_bed_path = temporary_path('.bed')
            # From BED to SQL #
            track.convert(orig_bed_path, test_sql_path, assembly='sacCer2')
            self.assertTrue(assert_file_equal(orig_sql_path, test_sql_path))
            # From SQL to BED #
            with track.load(test_sql_path) as t: t.roman_to_integer()
            track.convert(test_sql_path, test_bed_path)
            self.assertTrue(assert_file_equal(orig_bed_path, test_bed_path, start_b=1))
            # Clean up #
            os.remove(test_sql_path)
            os.remove(test_bed_path)

#-----------------------------------#
# This code was written by the BBCF #
# http://bbcf.epfl.ch/              #
# webmaster.bbcf@epfl.ch            #
#-----------------------------------#



"""
Contains the compressed file tests.
"""

# Built-in modules #
import os

# Internal modules #
import track
from track.common import temporary_path, assert_file_equal
from track.test import samples

# Unittesting module #
try:
    import unittest2 as unittest
except ImportError:
    import unittest

# Nosetest flag #
__test__ = True

###################################################################################
class TestConversion(unittest.TestCase):
    def runTest(self):
        # Prepare paths #
        orig_gzip_path = samples['gzip_tracks'][1]['gzip']
        orig_sql_path  = samples['small_features'][1]['sql']
        test_sql_path  = temporary_path('.sql')
        # From BED to SQL #
        track.convert(orig_gzip_path, test_sql_path, assembly='sacCer2')
        self.assertTrue(assert_file_equal(orig_sql_path, test_sql_path))
        # Clean up #
        os.remove(test_sql_path)

class TestWithoutExtension(unittest.TestCase):
    def runTest(self):
        in_path = samples['gzip_tracks'][4]['gzip']
        with track.load(in_path) as t:
            data = t.read()
            got = list(data)
        expected = [('chr1', 10, 20, u'Lorem', 1.0, 1),
                    ('chr1', 30, 40, u'Ipsum', 2.0, 1),
                    ('chr2', 10, 20, u'Dolor', 3.0, 1)]
        self.assertEqual(got, expected)



"""
Contains format conversion tests.
"""

# Built-in modules #
import os

# Internal modules #
import track
from track.test import samples
from track.common import temporary_path, assert_file_equal

# Unittesting module #
try:
    import unittest2 as unittest
except ImportError:
    import unittest

# Nosetest flag #
__test__ = True

###################################################################################
class TestSQLtoBED(unittest.TestCase):
    """Convert a file to BED"""
    def runTest(self):
        in_path = samples['small_features'][1]['sql']
        out_path = temporary_path('.bed')
        track.convert(in_path, out_path)
        #assert_file_equal(out_path, samples['small_features'][1]['bed'])
        os.remove(out_path)

#-----------------------------------#
# This code was written by the BBCF #
# http://bbcf.epfl.ch/              #
# webmaster.bbcf@epfl.ch            #
#-----------------------------------#



# coding: utf-8

################################### Modules ####################################
import sys
from track import common
from track.test import samples

################################# Description ##################################
label           = 'merge_scores'
short_name      = 'Merge scores'
long_name       = 'Merge scores of all signals together'
input_tracks    = [{'key':'n_tracks', 'position':1, 'kind': 'many', 'no-overlap':True,
                    'fields':['start','end','score']}]
input_args      = [{'key':'geometric', 'position':2, 'type': bool, 'default': False,
                    'doc':'Use the geometric mean instead of the arithmetic mean.'}]
input_meta      = []
output_tracks   = [{'position':1, 'fields': ['start','end','score'], 'datatype':'signal'}]
tracks_collapse = None
fields_collapse = None
chroms_collapse = 'union'

################################ Documentation #################################
tooltip = \
"""
The ``merge_scores`` manipulation merges N signals
using some average function. If the boolean value ``geometric``
is true, the geometric mean is used, otherwise the arithmetic
mean is used.
"""

tickets = ['https://github.com/bbcf/gMiner/issues/2']

################################## Examples ####################################
numeric_example = \
"""
X1 [start,end,score]: (10,20,50) (30,40,100)
X2 [start,end,score]: (10,12,20)
R  [start,end,score]: (10,20,40) (12,20,25) (30,40,50)
"""

visual_example = \
u"""
X1: ▁▁▁▁▁▁▁▁▁▁█████████▁▁▁▁▁▁
X2: ▁▁▁▁▁▅▅▅▅▅▅▅▅▅▅▁▁▁▁▁▁▁▁▁▁
R:  ▁▁▁▁▁▂▂▂▂▂▇▇▇▇▇▅▅▅▅▁▁▁▁▁▁
"""

#################################### Tests #####################################
tests = [
    {'tracks':   {'n_tracks': [samples['small_signals'][1]['sql'],
                               samples['small_signals'][2]['sql'],
                               samples['small_signals'][3]['sql']]},
     'args':     {'l':200},
     'chromosome': 'chrI',
     'expected': [( 0,    5,    2.0 + 0.6666666666666666),
                  ( 5,   10,    4.0),
                  ( 20,  30,   10.0),
                  ( 30,  40,   30.0),
                  ( 40,  50,   26.0 + 0.666666666666666),
                  ( 50,  60,  120.0),
                  ( 60,  68,  100.0),
                  ( 68,  70,  200.0),
                  ( 70,  80,  100.0),
                  ( 90, 110,    3.0),
                  (120, 130,   10.0)]}]

############################### Implementation #################################
def generate(n_tracks, geometric=False):
    # Get all iterators #
    sentinel = (sys.maxint, sys.maxint, 0.0)
    tracks = [common.sentinelize(x, sentinel) for x in n_tracks]
    elements = [x.next() for x in tracks]
    tracks_denom = 1.0/len(tracks)
    # Choose meaning function #
    if geometric: mean_fn = lambda x: sum(x)**tracks_denom
    else:         mean_fn = lambda x: sum(x)*tracks_denom
    # Check empty #
    for i in xrange(len(tracks)-1, -1, -1):
        if elements[i] == sentinel:
            tracks.pop(i)
            elements.pop(i)
    # Core loop #
    while tracks:
        # Find the next boundaries #
        start = min([x[0] for x in elements])
        end = min([x[0] for x in elements if x[0] > start] + [x[1] for x in elements])
        # Scores between boundaries #
        scores = [x[2] for x in elements if x[1] > start and x[0] < end]
        if scores: yield (start, end, mean_fn(scores))
        # Iterate over elements #
        for i in xrange(len(tracks)-1, -1, -1):
            # Change all starts #
            if elements[i][0] < end:
                elements[i] = (end, elements[i][1], elements[i][2])
            # Advance the elements that need to be advanced #
            if elements[i][1] <= end:
                elements[i] = tracks[i].next()
            # Pop sentinels #
            if elements[i] == sentinel:
                tracks.pop(i)
                elements.pop(i)



"""
Common stuff for python projects.
"""

########################### FILES FUNCTIONS ###################################
def iterate_lines(path, comment_char="#", linebreak_char="\\"):
    """
    Iterate over the lines of a text file in an intelligent way.
    1) Empty lines are skipped.
    2) Lines ending with line break characters such as '\\' are assembled.
    3) Lines starting with comments characters such as '#' are skipped.
    4) If the file is a GZIP, it is decompressed on the fly.
    This function yields the line number and the line content as a tuple.
    """
    # The pieces of a file are every block of data separated by the newline character.
    # The units of the file are composed by taking the pieces and removing empty and comment lines.
    # The actual lines of the file are composed by taking the units and joining line breaks.
    def units(iterable):
        for number, piece in enumerate(iterable):
            piece = piece.strip()
            if len(piece) == 0: continue
            if piece.startswith(comment_char): continue
            yield number, piece
    def lines(iterable):
        for number, unit in iterable:
            while unit.endswith(linebreak_char):
                unit = unit.strip(linebreak_char)
                unit += iterable.next()[1]
            yield number, unit
    if is_gzip(path):
        import gzip
        file = gzip.open(path, 'r')
        for line in lines(units(file)): yield line
        file.close()
    else:
        with open(path, 'r') as file:
            for line in lines(units(file)): yield line

#------------------------------------------------------------------------------#
def make_file_names(path):
    """
    Given a path to a file, generate more filenames.

    ::

        >>> g = make_file_names("tmp/test.png")
        >>> g.next()
        'tmp/test.png'
        >>> g.next()
        'tmp/test_1.png'
        >>> g.next()
        'tmp/test_2.png'
    """
    yield path
    import sys, os
    constant, ext = os.path.splitext(path)
    for i in xrange(1,sys.maxint): yield constant + "_" + str(i) + ext

#------------------------------------------------------------------------------#
def check_path(path):
    """
    Raise an exception if the path *path* is already taken.
    """
    import os
    if os.path.exists(path): raise Exception("The location '" + path + "' is already taken.")

#------------------------------------------------------------------------------#
def check_file(path):
    """
    Raise an exception if the path *path* is empty.
    """
    import os
    if os.path.getsize(path) == 0: raise Exception("The file '" + path + "' is empty.")

#------------------------------------------------------------------------------#
def check_executable(tool_name):
    """Raises an exception if the executable *tool_name* is not found."""
    import subprocess
    try:
        subprocess.Popen([tool_name], stderr=subprocess.PIPE)
    except OSError:
        raise Exception("The executable '" + tool_name + "' cannot be found")

#------------------------------------------------------------------------------#
def run_tool(tool_name, args):
    """
    Run the executable *tool_name* with *args* as arguments.
    Raises an exception if the process outputs something on the
    standard error output.
    """
    import subprocess
    proc = subprocess.Popen([tool_name] + args, stderr=subprocess.PIPE)
    stdout, stderr = proc.communicate()
    if stderr: raise Exception("The tool '" + tool_name + "' exited with message: " + '"' + stderr.strip('\n') + '"')

#------------------------------------------------------------------------------#
def temporary_path(suffix=''):
    """
    Often, one needs a new random and temporary file path
    instead of the random and temporary file object provided
    by the 'tempfile' module.
    """
    import tempfile
    file = tempfile.NamedTemporaryFile(suffix=suffix)
    path = file.name
    file.close()
    return path

#------------------------------------------------------------------------------#
def if_url_then_get_url(path):
    """
    If the path is an URL, download the file at that URL
    into a temporary file and return the path to that temporary
    file.
    If the path isn't an URL, return the path unchanged.
    """
    import os, urllib
    if path.startswith('http://') or path.startswith('https://'):
        extension = os.path.splitext(path)[1]
        tmp_path = temporary_path(extension)
        urllib.urlretrieve(path, tmp_path)
        return tmp_path
    else: return path

#------------------------------------------------------------------------------#
def is_gzip(path):
    """
    Returns true if the file at the given path is
    compressed using the gzip compression format.
    """
    with open(path, 'r') as file:
        if file.read(2) == '\x1f\x8b': return True
    return False

#------------------------------------------------------------------------------#
def load_dump(path, kind=None):
    if not kind:
        with open(path, 'r') as f: header = f.read(15)
        if header == 'SQLite format 3': kind = 'sql'
        else:                           kind = 'txt'
    if kind == 'txt':
        return open(path)
    if kind == 'sql':
        import sqlite3
        return sqlite3.connect(path).iterdump()

def assert_file_equal(pathA, pathB, start_a=0, start_b=0, end=999999, difflen=40, showlen=10):
    """
    Compare two file. If they are text they are simply loaded.
    If they are sqlite3 databases they are compared via their dumps
    Raise an exception if they are not equal, otherwise returns True.

    :param start_a: Start after these many lines from pathA.
    :param start_b: Start after these many lines from pathB.
    :param end: Stop after these many lines.
    :param difflen: Compute the difference on these many lines.
    :param showlen: Show these many lines from the difference.
    """
    # Load them #
    A, B = load_dump(pathA), load_dump(pathB)
    # Advance them #
    try:
        for x in xrange(start_a): A.next()
        for x in xrange(start_b): B.next()
    except StopIteration:
        pass
    # Compare them #
    import itertools
    for a,b,i in itertools.izip_longest(A,B,xrange(end)):
        if a != b:
            # They are different #
            import difflib
            diff = difflib.ndiff([a] + list(A)[0:difflen], [b] + list(B)[0:difflen])
            diff_text = '\n'.join(get_n_items_or_less(showlen, diff))
            # Raise and print differences #
            message = "The files:\n   %s'%s'%s and\n   %s'%s'%s differ.\n\n"
            message = message  % (Color.cyn, pathA, Color.end, Color.cyn, pathB, Color.end)
            message += "First %i lines of %i lines difference follows:\n"
            message = message  % (showlen, difflen)
            raise AssertionError(message + Color.wht + diff_text + Color.end)
    # They are identical #
    return True

#------------------------------------------------------------------------------#
def empty_sql_file(path):
    """
    Create an empty sql file at the path specified.
    """
    import sqlite3
    connection = sqlite3.connect(path)
    cursor = connection.cursor()
    connection.commit()
    cursor.close()
    connection.close()

#------------------------------------------------------------------------------#
def empty_file(path):
    """
    Create an empty file at the path specified.
    """
    with open(path, 'w') as f: pass

########################## ITERATOR FUNCTIONS ##################################
def sentinelize(iterable, sentinel):
    """
    Add an item to the end of an iterable.

    ::

        >>> list(sentinelize(range(4), 99))
        [0, 1, 2, 3, 99]
    """
    for item in iterable: yield item
    yield sentinel

#------------------------------------------------------------------------------#
def get_next_item(iterable):
    """
    Gets the next item of an iterable.
    If the iteratable is exhausted, returns None.

    ::

        >>> get_next_item(iter(range(4)))
        0
        >>> get_next_item(iter([]))
    """
    try:
        x = iterable.next()
    except StopIteration:
        x = None
    except AttributeError:
        x = None
    return x

#------------------------------------------------------------------------------#
def get_n_items_or_less(n, iterable):
    """
    Gets the n next items of an iterable or less
    if the iterator is exhausted before

    ::

        >>> get_n_items_or_less(6, iter(range(10)))
        [0, 1, 2, 3, 4, 5]
        >>> get_n_items_or_less(6, iter(range(4)))
        [0, 1, 2, 3]
    """
    result = []
    try:
        for i in xrange(n):
            result.append(iterable.next())
    except StopIteration:
        pass
    return result

#------------------------------------------------------------------------------#
def pick_iterator_elements(iterable, indices):
    """
    Return a new iterator, yielding only those elements
    of the original iterator.

    ::

        >>> list(pick_iterator_elements(iter([('a','b','c'),(10,20,30)]), (0,2)))
        [['a', 'c'], [10, 30]]
    """
    for item in iterable: yield [item[i] for i in indices]

############################## VARIOUS FUNCTIONS ###############################
def format_float(f, precision=None):
    """
    Formats a float, by rounding it and striping unnecesarry zeros.
    But it always leaves one zero after the decimal point.

    ::

        >>> format_float(0.100)
        '0.1'
        >>> format_float(0.1001)
        '0.1001'
        >>> format_float(0.10000001)
        '0.1'
        >>> format_float(101.0)
        '101.0'
        >>> format_float(100001.1)
        '100000.0'
        >>> format_float(1000.0/6.0)
        '166.7'
        >>> format_float(999.9)
        '999.9'
        >>> format_float(999.99)
        '1000.0'
    """
    if precision: return '%s' % float('%.'+str(precision)+'g' % f)
    else:         return '%s' % float('%.4g' % f)

#------------------------------------------------------------------------------#
def natural_sort(item):
    """
    Sort strings that contain numbers correctly.

    ::

        >>> l = ['v1.3.12', 'v1.3.3', 'v1.2.5', 'v1.2.15', 'v1.2.3', 'v1.2.1']
        >>> l.sort(key=natural_sort)
        >>> l.__repr__()
        "['v1.2.1', 'v1.2.3', 'v1.2.5', 'v1.2.15', 'v1.3.3', 'v1.3.12']"
    """
    if item is None: return 0
    import re
    def try_int(s):
        try: return int(s)
        except ValueError: return s
    return map(try_int, re.findall(r'(\d+|\D+)', item))

#------------------------------------------------------------------------------#
def andify_strings(list_of_strings):
    """
    Given a list of strings will join them with commas
    and a final "and" word.

    >>> andify_strings(['Apples', 'Oranges', 'Mangos'])
    'Apples, Oranges and Mangos'
    """
    result = ', '.join(list_of_strings)
    comma_index = result.rfind(',')
    if comma_index > -1: result = result[:comma_index] + ' and' + result[comma_index+1:]
    return result

#------------------------------------------------------------------------------#
def collapse(method, collection):
    """
    Collapse lists in specific ways.

    :param method: the method used to collapse. Can be one of: ``adding``, ``appending``, ``union`` or ``intersection``,
    :type method: string
    :param collection: a list of things
    :type collection: list
    :returns: a new list

    >>> collapse('adding',[1,5,2])
    8
    >>> collapse('adding',[3])
    3
    >>> collapse('appending',[[1,5],[5,3],[2,1]])
    [1, 5, 5, 3, 2, 1]
    >>> collapse('appending',[[1,2,3]])
    [1, 2, 3]
    >>> collapse('union',[[1,5],[5,3],[2,1]])
    [1, 5, 3, 2]
    >>> collapse('intersection',[[1,5,4],[5,3,3],[2,6,5]])
    [5]
    >>> collapse('intersection',[['a', 'b', 'c', 'd'], ['a', 'b', 'c', 'd', 'e', 'f']])
    ['a', 'b', 'c', 'd']
    >>> collapse('first',[[1,5,4],[5,3,3],[2,6,5]])
    [1, 5, 4]
    >>> collapse('last',[[1,5,4],[5,3,3],[2,6,5]])
    [2, 6, 5]
    """
    if method == 'adding':
        return sum(collection)
    elif method == 'appending':
        return [x for y in collection for x in y]
    elif method == 'union':
        seen = set()
        return [x for y in collection for x in y if x not in seen and not seen.add(x)]
    elif method == 'intersection':
        common_items = reduce(set.intersection, map(set,collection))
        return [i for i in collection[0] if i in common_items]
    elif method == 'first':
        return collection[0]
    elif method == 'last':
        return collection[-1]

#------------------------------------------------------------------------------#
def int_to_roman(input):
    """
    Convert an integer to a roman numeral.

    ::

        >>> int_to_roman(1999)
        'MCMXCIX'
    """
    ints = (1000, 900,  500, 400, 100,  90, 50,  40, 10,  9,   5,  4,   1)
    nums = ('M',  'CM', 'D', 'CD', 'C', 'XC', 'L', 'XL', 'X', 'IX', 'V', 'IV', 'I')
    if type(input) != type(1): raise TypeError, 'Expected integer, got "%s."' % type(input)
    if not 0 < input < 4000: raise ValueError, 'Argument must be between 1 and 3999.'
    result = ""
    for i in range(len(ints)):
       count   = int(input / ints[i])
       result += nums[i] * count
       input  -= ints[i] * count
    return result

#------------------------------------------------------------------------------#
def roman_to_int(input):
    """
    Convert a roman numeral to an integer.

    >>> orig_integers = range(1, 4000)
    >>> romans = [int_to_roman(i) for i in orig_integers]
    >>> computed_integers = [roman_to_int(r) for r in romans]
    >>> print orig_integers == computed_integers
    True
    """
    ints = [1000, 500, 100, 50,  10,  5,   1]
    nums = ['M', 'D', 'C', 'L', 'X', 'V', 'I']
    if type(input) != type(""): raise TypeError, 'Expected string, got "%s."' % type(input)
    input = input.upper()
    places = []
    for c in input:
        if not c in nums:
            raise ValueError, 'Input is not a valid roman numeral: "%s."' % input
    for i in range(len(input)):
        c = input[i]
        value = ints[nums.index(c)]
        try:
            nextvalue = ints[nums.index(input[i +1])]
            if nextvalue > value: value *= -1
        except IndexError: pass
        places.append(value)
    output = sum([n for n in places])
    if int_to_roman(output) == input: return output
    else: raise ValueError, 'Input is not a valid roman numeral: "%s."' % input

#------------------------------------------------------------------------------#
def profile_it(command):
    """Profiles a given command and returns a pstat object.
       Use it like this:

            r = profile_it('batch_benchmark(args)')
            r.sort_stats('time', 'cum')
            r.print_stats(10)

    """
    import os, cProfile, pstats
    result_file = temporary_path('.profile')
    cProfile.run(command, result_file)
    results = pstats.Stats(result_file)
    results.sort_stats('time', 'cum')
    results.print_stats(20)
    os.remove(result_file)
    return results

#------------------------------------------------------------------------------#
def import_module(name):
    """
    Import a module that given it's name
    and return that module.
    """
    __import__(name)
    import sys
    return sys.modules[name]

############################## CLASSES #########################################
class Color:
    """Shortcuts for the ANSI escape sequences to control
       formatting, color, etc. on text terminals. Use it like this::

            print Color.red + "Hello world" + Color.end
    """
    # Special #
    end = '\033[0m'
    # Regular #
    blk   = '\033[0;30m' # Black
    red   = '\033[0;31m' # Red
    grn   = '\033[0;32m' # Green
    ylw   = '\033[0;33m' # Yellow
    blu   = '\033[0;34m' # Blue
    pur   = '\033[0;35m' # Purple
    cyn   = '\033[0;36m' # Cyan
    wht   = '\033[0;37m' # White
    # Bold #
    bold  = '\033[1m'
    b_blk = '\033[1;30m' # Black
    b_red = '\033[1;31m' # Red
    b_grn = '\033[1;32m' # Green
    b_ylw = '\033[1;33m' # Yellow
    b_blu = '\033[1;34m' # Blue
    b_pur = '\033[1;35m' # Purple
    b_cyn = '\033[1;36m' # Cyan
    b_wht = '\033[1;37m' # White
    # Light #
    light = '\033[2m'
    l_blk = '\033[2;30m' # Black
    l_red = '\033[2;31m' # Red
    l_grn = '\033[2;32m' # Green
    l_ylw = '\033[2;33m' # Yellow
    l_blu = '\033[2;34m' # Blue
    l_pur = '\033[2;35m' # Purple
    l_cyn = '\033[2;36m' # Cyan
    l_wht = '\033[2;37m' # White
    # Italic #
    italic = '\033[1m'
    i_blk = '\033[3;30m' # Black
    i_red = '\033[3;31m' # Red
    i_grn = '\033[3;32m' # Green
    i_ylw = '\033[3;33m' # Yellow
    i_blu = '\033[3;34m' # Blue
    i_pur = '\033[3;35m' # Purple
    i_cyn = '\033[3;36m' # Cyan
    i_wht = '\033[3;37m' # White
    # Underline #
    underline = '\033[4m'
    u_blk = '\033[4;30m' # Black
    u_red = '\033[4;31m' # Red
    u_grn = '\033[4;32m' # Green
    u_ylw = '\033[4;33m' # Yellow
    u_blu = '\033[4;34m' # Blue
    u_pur = '\033[4;35m' # Purple
    u_cyn = '\033[4;36m' # Cyan
    u_wht = '\033[4;37m' # White
    # Glitter #
    flash = '\033[5m'
    g_blk = '\033[5;30m' # Black
    g_red = '\033[5;31m' # Red
    g_grn = '\033[5;32m' # Green
    g_ylw = '\033[5;33m' # Yellow
    g_blu = '\033[5;34m' # Blue
    g_pur = '\033[5;35m' # Purple
    g_cyn = '\033[5;36m' # Cyan
    g_wht = '\033[5;37m' # White
    # Fill #
    f_blk = '\033[40m'   # Black
    f_red = '\033[41m'   # Red
    f_grn = '\033[42m'   # Green
    f_ylw = '\033[43m'   # Yellow
    f_blu = '\033[44m'   # Blue
    f_pur = '\033[45m'   # Purple
    f_cyn = '\033[46m'   # Cyan
    f_wht = '\033[47m'   # White

#------------------------------------------------------------------------------#
class JournaledDict(object):
    """
    A dictionary like object that tracks modification to itself.
    It has an extra special boolean self.modified value that starts off with "False"
    It has three special methods: self.save(), self.rollback() and self.overwrite()

       * rollback() will reset the self.modified variable to False and restore
    the dictionary to it's previous state.

       * save()  will reset the self.modified variable to False and remember
    the dictionary current's state.

       * overwrite(d) will reset the self.modified variable to True and repalce
    the dictionary with the one passed to it.
    """

    def __init__(self, d=None):
        # The dictionary itself #
        if d: self.data = d
        else: self.data = {}
        # The remembered state #
        self.backup = self.data.copy()
        # The boolean value #
        self.modified = False

    def __cmp__(self, d):
        if isinstance(d, JournaledDict): return cmp(self.data, d.data)
        else: return cmp(self.data, d)

    def __getitem__(self, key):
        if key in self.data: return self.data[key]
        if hasattr(self.__class__, "__missing__"): return getattr(self.__class__, "__missing__")(self, key)
        raise KeyError(key)

    def __setitem__(self, key, item):
        self.modified = True
        self.data[key] = item

    def __delitem__(self, key):
        self.modified = True
        del self.data[key]

    def __repr__(self): return repr(self.data)
    def __iter__(self): return iter(self.data)
    def __contains__(self, key): return key in self.data
    def __len__(self): return len(self.data)
    def keys(self): return self.data.keys()
    def items(self): return self.data.items()
    def iteritems(self): return self.data.iteritems()
    def iterkeys(self): return self.data.iterkeys()
    def itervalues(self): return self.data.itervalues()
    def values(self): return self.data.values()
    def has_key(self, key): return key in self.data
    def get(self, key, d=None): return self.data.get(key, d)

    def clear(self):
        self.modified = True
        self.data.clear()

    def setdefault(self, key, failobj=None):
        if key not in self:
            self.modified = True
            self[key] = failobj
        return self[key]

    def pop(self, key, *args):
        self.modified = True
        return self.data.pop(key, *args)

    def popitem(self):
        self.modified = True
        return self.data.popitem()

    def copy(self):
        return JournaledDict(self.data.copy())

    def update(self, d=None, **kwargs):
        if d is None: return
        self.modified = True
        if isinstance(d, JournaledDict): self.data.update(d.data)
        elif isinstance(d, dict) or not hasattr(d, 'items'): self.data.update(d)
        else:
            for k, v in d.items(): self[k] = v
        if len(kwargs): self.data.update(kwargs)

    @classmethod
    def fromkeys(cls, iterable, value=None):
        d = cls()
        for key in iterable: d[key] = value
        return d

    def save(self):
        self.modified = False
        self.backup = self.data.copy()

    def rollback(self):
        self.modified = False
        self.data = self.backup.copy()

    def overwrite(self, d):
        self.modified = True
        self.data = d

#------------------------------------------------------------------------------#
class Timer(object):
    """Time a given code block. Use it like this::

            with Timer() as t: cool_benchmark(args)
            print t.total_time
    """

    def __init__(self):
        import timeit
        self.timer = timeit.default_timer

    def __enter__(self):
        self.start = self.timer()
        return self

    def __exit__(self, *args):
        self.end = self.timer()
        self.total_time = self.end - self.start

#------------------------------------------------------------------------------#
class Path(object):
    """
    Easy access to the to functions provided by the os.path module
    Use it like this::

        path = path("tmp/images/lena.png")
        print path.extension
        print path.filename
        print path.directory
    """

    def __init__(self, path):
        import os
        self.path = path
        self.extension = os.path.splitext(path)[0]
        self.directory = os.path.dirname(path) + '/'
        self.filename = os.path.splitext(os.path.basename(path))[0]

    def __repr__(self):
        return self.path

#------------------------------------------------------------------------------#
class Filesize(object):
    """
    Container for a size in bytes with a human readable representation
    Use it like this::

        >>> size = Filesize(123123123)
        >>> print size
        117.4 MB
    """

    chunk = 1024
    units = ['bytes', 'KB', 'MB', 'GB', 'TB', 'PB']
    precisions = [0, 0, 1, 2, 2, 2]

    def __init__(self, value):
        import os
        if isinstance(value, int): self.size = value
        else: self.size = os.path.getsize(str(value))

    def __int__(self):
        return self.size

    def __float__(self):
        return float(int(self))

    def __repr__(self):
        if self.size == 0: return '0 bytes'
        from math import log
        unit = self.units[min(int(log(self.size, self.chunk)), len(self.units) - 1)]
        return self.format(unit)

    def format(self, unit):
        if unit not in self.units: raise Exception("Not a valid file size unit: %s" % unit)
        if self.size == 1 and unit == 'bytes': return '1 byte'
        exponent = self.units.index(unit)
        quotient = float(self.size) / self.chunk**exponent
        precision = self.precisions[exponent]
        format_string = '{0:.%sf} {1}' % (precision)
        return format_string.format(quotient, unit)



"""
Provides easy access to some meta data about genome assemblies. You can retrieve the list of chromosomes for any assembly. In addition, you can get the lengths in base pairs of every chromosome.
"""

# Built-in modules #
import os, re, sqlite3

# The database #
path = os.path.abspath(os.path.dirname(__file__)) + '/genomes.db'
connection = sqlite3.connect(path)
cursor = connection.cursor()

# The list of assemblies #
cursor.execute("select name from sqlite_master where type='table'")
assemblies = [x[0] for x in cursor if x[0] != 'assemblies']

################################################################################
class Assembly(object):
    """The only object provided by the library.

       :param assembly: A valid assembly name such as 'sacCer2'.
       :type  assembly: string
    """

    def __init__(self, assembly):
        # Record the name #
        self.name = assembly
        # Check the input type #
        if not isinstance(assembly, basestring):
            raise TypeError('The assembly paramater needs to be a string such as "sacCer2".')
        if not assembly in assemblies:
            raise TypeError('The assembly "%s" was not found in the database.' % assembly)
        # Load the chromosome data #
        cursor.execute("select * from '%s'" % assembly)
        columns = [x[0] for x in cursor.description]
        self.chromosomes = [dict(zip(columns,chrom)) for chrom in cursor]
        # Cut the synonyms #
        for chrom in self.chromosomes:
            chrom['synonyms'] = chrom['synonyms'].split(',') if chrom['synonyms'] else []

    def __repr__(self): return '<%s object "%s">' % (self.__class__.__name__, self.name)

    @property
    def chrmeta(self):
        """A dictionary of chromosome metadata:

        ::

            >>> from genomes import Assembly
            >>> a = Assembly('TAIR10')
            >>> print a.chrmeta
            {u'c': {'length': 154478, 'refseq': u'NC_000932.1'}, u'm': {'length': 366924, 'refseq': u'NC_001284.2'}, u'1': {'length': 30427671, 'refseq': u'NC_003070.9'}, u'3': {'length': 23459830, 'refseq': u'NC_003074.8'}, u'2': {'length': 19698289, 'refseq': u'NC_003071.7'}, u'5': {'length': 26975502, 'refseq': u'NC_003076.8'}, u'4': {'length': 18585056, 'refseq': u'NC_003075.7'}}
        """
        result = {}
        for chrom in self.chromosomes:
            chrom_info = {}
            chrom_info['length'] = chrom['length']
            chrom_info['refseq'] = '%s.%s' % (chrom['refseq_locus'], chrom['refseq_version'])
            result[chrom['label']] = chrom_info
        return result

    def guess_chromosome_name(self, chromosome_name):
        """Searches the assembly for chromosome synonym names, and returns the canonical name of the chromosome.

           :param chromosome_name: Any given name for a chromosome in this assembly.
           :type  chromosome_name: string

           :returns: The same or an other name for the chromosome. Returns None if the chromosome is not known about.

           ::

               >>> from genomes import Assembly
               >>> a = Assembly('sacCer2')
               >>> print a.guess_chromosome_name('chrR')
               2micron
               >>> a = Assembly('hg19')
               >>> print a.guess_chromosome_name('NC_000023.9')
               chrX
        """
        # Convert to unicode #
        chromosome_name = unicode(chromosome_name)
        # Check for synonyms #
        for chrom in self.chromosomes:
            if chromosome_name in chrom['synonyms']: return chrom['label']
        # Check for NCBI nomenclature #
        if chromosome_name.startswith('NC_'):
            name = re.sub('\.[0-9]+$', '', chromosome_name)
            for chrom in self.chromosomes:
                if name == chrom['refseq_locus']: return chrom['label']
        # Do some guessing #
        name = chromosome_name.lstrip('chr')
        for chrom in self.chromosomes:
            if name == chrom['name']: return chrom['label']
            if name == chrom['label'].strip('chr'): return chrom['label']

################################################################################
if __name__ == "__main__":
    import doctest
    doctest.testmod()



"""
This module implements the parsing of GFF files.

http://genome.ucsc.edu/FAQ/FAQformat.html#format3

A bug in the python sqlite3 module prevents us form using 'group' for the ninth field.
http://bugs.python.org/issue9750

It is hence replaced with 'attributes'
"""

# Built-in modules #
import shlex

# Internal modules #
from track.parse import Parser
from track.common import iterate_lines
from track.util import strand_to_int

# Constants #
all_fields = ['source', 'name', 'start', 'end', 'score', 'strand', 'frame', 'attributes']

################################################################################
class ParserGFF(Parser):
    format = 'gff'
    def parse(self):
        # Initial variables #
        fields = []
        info   = {}
        # Main loop #
        for number, line in iterate_lines(self.path):
            # Ignored lines #
            if line.startswith("browser "): continue
            # Track headers #
            if line.startswith("track "):
                try:
                    info = dict([p.split('=',1) for p in shlex.split(line[6:])])
                except ValueError:
                    self.handler.error("The track%s seems to have an invalid <track> header line", self.path, number)
                fields = []
                continue
            # Split the lines #
            items = line.split('\t')
            if len(items) == 1: items = line.split()
            # Chromosome #
            chrom = items.pop(0)
            # Length is nine #
            if len(items) != 8:
                self.handler.error("The track%s doesn't have nine columns", self.path, number)
            # Have we started a track already ? #
            if not fields:
                self.handler.newTrack(info, self.name)
                fields = all_fields[0:len(items)]
                self.handler.defineFields(fields)
            # Source field #
            if items[0] == '.': items[0] = ''
            # Name field #
            if items[1] == '.': items[1] = ''
            # Start and end fields #
            try:
                items[2] = int(items[2])
                items[3] = int(items[3])
            except ValueError:
                self.handler.error("The track%s has non integers as interval bounds", self.path, number)
            if items[3] <= items[2]:
                self.handler.error("The track%s has negative or null intervals", self.path, number)
            # Score field #
            if items[4] == '.' or items[4] == '': items[4] = 0.0
            try:
                items[4] = float(items[4])
            except ValueError:
                self.handler.error("The track%s has non floats as score values", self.path, number)
            # Strand field #
            items[5] = strand_to_int(items[5])
            # Frame field #
            if items[6] == '.': items[6] = None
            else:
                try:
                    items[6] = int(items[6])
                except ValueError:
                    self.handler.error("The track%s has non integers as frame value", self.path, number)
            # Group or attribute field #
            if items[7] == '.': items[7] = ''
            # Yield it #
            self.handler.newFeature(chrom, items)

#-----------------------------------#
# This code was written by the BBCF #
# http://bbcf.epfl.ch/              #
# webmaster.bbcf@epfl.ch            #
#-----------------------------------#



"""
This module implements the parsing of SGA files.

http://ccg.vital-it.ch/chipseq/doc/chipseq_tech.php
"""

# Internal modules #
from track.parse import Parser
from track.util import strand_to_int
from track.common import iterate_lines

# Constants #
all_fields = ['name', 'start', 'end', 'strand', 'score']

################################################################################
class ParserSGA(Parser):
    format = 'sga'

    def parse(self):
        # Initial variables #
        self.handler.newTrack({'int_to_float':'score'}, self.name)
        self.handler.defineFields(all_fields)
        with open(self.path) as f:
            for number, line in enumerate(f):
                items = line.split('\t')
                chrom = items.pop(0)
                if len(items) < 4: self.handler.error("The track%s doesn't have five columns", self.path, number)
                # Name field #
                name = items[0]
                # Start field #
                try: pos = int(items[1])
                except ValueError: self.handler.error("The track%s has non integers as position", self.path, number)
                # Strand field #
                strand = strand_to_int(items[2])
                # Score field #
                try: score = int(items[3])
                except ValueError: self.handler.error("The track%s has non integers as tag count values", self.path, number)
                # Yield it #
                self.handler.newFeature(chrom, (name, pos-1, pos, strand, score))

    def parse(self):
        # Initial variables #
        l_chrom, l_name, l_start, l_end, l_strand, l_score = None, None, None, None, None, None
        # Start a new track #
        self.handler.newTrack({'int_to_float':'score'}, self.name)
        self.handler.defineFields(all_fields)
        # Line loop #
        for number, line in iterate_lines(self.path):
            # Ignored lines #
            if line.startswith("browser "): continue
            if line.startswith("track "): continue
            # Split the lines #
            items = line.split('\t')
            if len(items) == 1: items = line.split()
            # Is it a legal line ? #
            if len(items) < 5: self.handler.error("The track%s has less than five columns", self.path, number)
            # Chromosome #
            chrom = items.pop(0)
            # Name field #
            name = items[0]
            # Start and end field #
            try: pos = int(items[1])
            except ValueError: self.handler.error("The track%s has non integers as position", self.path, number)
            start = pos-1
            end = pos
            # Strand field #
            strand = strand_to_int(items[2])
            # Score field #
            try: score = int(items[3])
            except ValueError: self.handler.error("The track%s has non integers as tag count values", self.path, number)
            # Ignore null scores #
            if score == 0: continue
            # Merge adjacent features with same scores #
            if (l_chrom, l_name, l_strand, l_score) == (chrom, name, strand, score) and start == l_end:
                l_end = end
                continue
            else:
                if l_chrom: self.handler.newFeature(l_chrom, (l_name, l_start, l_end, l_strand, l_score))
                l_chrom, l_name, l_start, l_end, l_strand, l_score = chrom, name, start, end, strand, score
        # Last feature #
        if l_chrom: self.handler.newFeature(l_chrom, (l_name, l_start, l_end, l_strand, l_score))

#-----------------------------------#
# This code was written by the BBCF #
# http://bbcf.epfl.ch/              #
# webmaster.bbcf@epfl.ch            #
#-----------------------------------#



b'This module needs Python 2.6 or later.'

# Special variables #
__version__ = '1.1.0'
__all__ = ['load', 'new', 'convert']

# Other variables #
formats = ('sql', 'bed', 'wig', 'gff', 'gtf', 'bedgraph', 'bigwig', 'gzip', 'sga')

# Built-in modules #
import os, re, sqlite3, json
from itertools import imap

# Internal modules #
from track.parse import get_parser
from track.serialize import get_serializer
from track.util import determine_format, join_read_queries, make_cond_from_sel, parse_chr_file
from track.util import sql_field_types, py_field_types, serialize_chr_file
from track.util import gzip_inner_format
from track.common import check_path, check_file, empty_file, empty_sql_file, temporary_path
from track.common import JournaledDict, natural_sort, int_to_roman, roman_to_int
from track.common import Color, pick_iterator_elements, get_next_item, is_gzip
from track.common import if_url_then_get_url
from track.genomes import Assembly

# Compiled modules #
from track.pyrow import SuperRow

# Constants #
special_tables = ('attributes', 'chrNames', 'types')
minimum_fields = ('start', 'end')
default_fields = ('start', 'end', 'name', 'score', 'strand')
signal_fields = ('start', 'end', 'score')
feature_fields = ('start', 'end', 'name', 'score', 'strand', 'attributes')
relational_fields = ('start', 'end', 'name', 'score', 'strand', 'attributes', 'group', 'id')

################################################################################
def load(path, format=None, readonly=False):
    """Loads a track from disk, whatever the format is.

       :param path: is the path to track file to load or an URL. If the path is an URL, the file will be downloaded automatically. If the path is a GZIP file, it will be decompressed automatically.
       :type  path: string
       :param format: is an optional parameter specifying the format of the track to load when it cannot be guessed from the file extension.
       :type  format: string
       :param readonly: is an optional parameter that defaults to ``False``. When set to ``True``, any operation attempting to write to the track will silently be ignored
       :type  readonly: bool
       :returns: a Track instance

       ::

            import track
            with track.load('tracks/rp_genes.bed') as rpgenes:
                data = rpgenes.read()
            with track.load('/tmp/ae456f0', 'sql') as t:
                data = t.read()
            with track.load('tracks/repeats.bed', readonly=True) as repeats:
                data = repeats.read()
            with track.load('http://example.com/genes.bed') as genes:
                data = genes.read()
    """
    # Check if URL #
    path = if_url_then_get_url(path)
    # Check not empty #
    check_file(path)
    # Guess the format #
    if not format: format = determine_format(path)
    # If sql, just make a track with the path #
    # Otherwise we need to convert the file #
    if format == 'sql':
        return Track(path, readonly)
    else:
        sql_path = temporary_path(".sql") or os.path.splitext(path)[0] + ".sql"
        convert(source=(path, format), destination=(sql_path, 'sql'))
        return Track(sql_path, readonly=readonly, orig_path=path, orig_format=format)

#---------------------------------------------------------------------------------#
def new(path, format=None):
    """Creates a new empty track in preparation for writing to it.

       :param path: is the path to track file to create.
       :type  path: string
       :param format: is an optional parameter specifying the format of the track to create when it cannot be guessed from the file extension.
       :type  format: string
       :returns: a Track instance

       ::

           import track
           with track.new('tmp/track.sql') as t:
               t.write('chr1', [(10, 20, 'Gene A', 0.0, 1)])
               t.set_chrmeta('hg19')
           with track.new('tracks/peaks.sql', 'sql') as t:
               t.fields = ['start', 'end', 'name', 'score']
               t.write('chr5', [(500, 1200, 'Peak1', 11.3)])
    """
    # Guess the format #
    if not format: format = os.path.splitext(path)[1][1:].lower()
    check_path(path)
    # If sql, just make a new track at the path #
    # Otherwise we need to make a temporary sql #
    if format == 'sql':
        empty_sql_file(path)
        return Track(path)
    else:
        sql_path = temporary_path(".sql") or os.path.splitext(path)[0] + ".sql"
        empty_file(path)
        empty_sql_file(sql_path)
        return Track(sql_path, orig_path=path, orig_format=format)

#---------------------------------------------------------------------------------#
def convert(source, destination, assembly=None):
    """Converts a track from one format to an other. The *source* file should have a different format from the *destination* file. If either the source or destination are missing a file extension, you can specify their formats using a tuple. See examples below.

       :param source: is the path to the original track to load.
       :type  source: string
       :param destination: is the path to the track to be created.
       :type  destination: string
       :param assembly: an optional compatible assembly name. Useful when the destination format needs to contain chromosome meta data and this is not available in the source file.
       :type  assembly: string

       :returns: the path to the track created (or a list of track paths in the case of multi-track files).

       ::

           import track
           track.convert('tracks/genes.bed', 'tracks/genes.sql')
           track.convert('tracks/genes.sql', 'tracks/genes.bigWig', assembly='hg19')
           track.convert(('tracks/no_extension', 'gff'), 'tracks/genes.sql')
           track.convert(('tmp/4afb0edf', 'bed'), ('tmp/converted', 'wig'))
    """
    # Parse the source parameter #
    if isinstance(source, tuple):
        source_path   = if_url_then_get_url(source[0])
        source_format = source[1]
    else:
        source_path   = if_url_then_get_url(source)
        source_format = determine_format(source_path)
    # Parse the destination parameter #
    if isinstance(destination, tuple):
        destination_path   = destination[0]
        destination_format = destination[1]
    else:
        destination_path   = destination
        destination_format = determine_format(destination_path)
    # Check it is not taken #
    check_path(destination_path)
    # Special cases #
    if destination_format == 'bigwig' and source_format != 'sql':
        source_path = convert((source_path, source_format), temporary_path('.sql'), assembly)
        source_format = 'sql'
    # Check it is not empty #
    check_file(source_path)
    # Check for compressed files #
    if is_gzip(source_path): source_format = gzip_inner_format(source_path)
    # Get a parser #
    parser = get_parser(source_path, source_format)
    # Get a serializer #
    serializer = get_serializer(destination_path, destination_format)
    # Tell the serializer about the assembly #
    if assembly: serializer.defineAssembly(assembly)
    # The serializer has a copy of the parser and vice-versa #
    serializer(parser)
    return parser(serializer)

################################################################################
class Track(object):
    """The track object itself is iterable and will yield the name of all chromosomes.
    ::

        import track
        with track.load('tracks/all_genes.sql') as t:
            for chrom in t: print chrom
            if 'chrY' in t: print 'Male'
            if len(t) != 23: print 'Aneuploidy'
    """

    def __init__(self, path, readonly=False, autosave=True, orig_path=None, orig_format=None):
        """The track package is designed to be accessed via the 'load()' and 'new()'
           functions in order to create Track objects.
           Usually, the constructor is not called directly."""
        # Passed attributes #
        self.path        = path
        self.readonly    = readonly
        self.autosave    = autosave
        self.orig_path   = orig_path
        self.orig_format = orig_format
        # Hidden attributes #
        self._modified = False
        self._fields   = []
        self._chrmeta  = JournaledDict()
        self._info     = JournaledDict()
        # Opening the database #
        self._connection = sqlite3.connect(self.path)
        self._connection.row_factory = SuperRow
        # A list to hold all cursors #
        self.all_cursors = []
        # Make two cursors #
        self._cursor       = self.cursor()
        self._write_cursor = self.cursor()
        # Load some tables #
        self._chrmeta_read()
        self._info_read()

    def __enter__(self):
        """Called when entering the 'with' statement."""
        return self

    def __exit__(self, errtype, value, traceback):
        """Called when exiting the 'with' statement.
        Enables us to close the database properly, even when exceptions are raised."""
        self.close()

    def __iter__(self):
        """Called when evaluating ``for chrom in t: pass``."""
        return iter(self.chromosomes)

    def __contains__(self, key):
        """Called when evaluating ``"chr1" in t``."""
        return key in self.chromosomes

    def __len__(self):
        """Called when evaluating ``len(t)``."""
        return len(self.chromosomes)

    def __nonzero__(self):
        """Called when evaluating ``if t: pass``."""
        return True

    def __getitem__(self, key):
        """Called when evaluating ``t[0] or t['chr1']``."""
        if isinstance(key,int): return self.chromosomes[key]
        else: return self.read(key)

    def __repr__(self):
        """Called when evaluating ``print t``."""
        return '<%s object on "%s">' % (self.__class__.__name__, self.path)

    #-----------------------------------------------------------------------------#
    @property
    def modified(self):
        """A boolean value which indicates if the track has been changed since it was opened. This value is set to False when you load a track and is set to True as soon, as you ``write``, ``rename`` or ``remove``. Changing the ``info`` or ``chrmeta`` attributes will also set this value to True."""
        if self._modified or self.info.modified or self.chrmeta.modified: return True
        return False

    @property
    def fields(self):
        """A list the value types that each feature in the track will contain. For instance:

               ``['start', 'end', 'name', 'score', 'strand']``

           Setting this attribute will influence the behaviour of all future read() and write() calls."""
        # Checks the user set self._fields attribute. If it is empty,
        # it gets the field names of the first chromosome table
        # it finds.
        if self._fields: return self._fields
        elif self.chromosomes: return self._get_fields_of_table(self.chromosomes[0])
        else: return []

    @fields.setter
    def fields(self, value):
        """Set the fields globally for the track. This value is then used by read() and write() to get the fields in the right order."""
        self._fields = value

    @property
    def tables(self):
        """The complete list of SQL tables."""
        self._cursor.execute("select name from sqlite_master where type='table'")
        return [x[0].encode('ascii') for x in self._cursor.fetchall()]

    @property
    def chromosomes(self):
        """A list of all available chromosome. For instance:

               ``['chr1, 'chr2', 'chr3', 'chr4', 'chr5', 'chrC', 'chrM']``

           You cannot set this attribute. To add new chromosomes, just ``write()`` to them."""
        # Filters the list of SQL tables to retrieve the list of chromosomes.
        chroms = [x for x in self.tables if x not in special_tables and not x.endswith('_idx')]
        chroms.sort(key=natural_sort)
        return chroms

    def _get_fields_of_table(self, chrom):
        """Return the list of fields for a particular table by querying the SQL for the complete list of column names"""
        # Check the table exists #
        if not chrom in self.tables: return []
        # A pragma statement will implicitly issue a commit, don't use #
        self._cursor.execute("SELECT * from '%s' LIMIT 1" % chrom)
        fields = [x[0] for x in self._cursor.description]
        self._cursor.fetchall()
        return fields

    #-----------------------------------------------------------------------------#
    def cursor(self):
        """Create a new sqlite3 cursor object connected to the track database. You can use this attribute to make your own SQL queries and fetch the results. More information is available on the `sqlite3 documentation pages <http://docs.python.org/library/sqlite3.html>`_.

        :returns: A new sqlite3 cursor object

        ::

            import track
            with track.load('tracks/rp_genes.sql') as rpgenes:
                cursor = rpgenes.cursor()
                cursor.execute("select name from sqlite_master where type='table'")
                results = cursor.fetchall()
        """
        new_cursor = self._connection.cursor()
        self.all_cursors.append(new_cursor)
        return new_cursor

    #-----------------------------------------------------------------------------#
    def save(self):
        """Store the changes that were applied to the track on the disk. If the track was loaded from a text file such as 'bed', the file is rewritten with the changes included. If the track was loaded as an SQL file, the changes are committed to the database. Calling ``rollback`` will revert all changes to the track since the last call to ``save()``. By default, when the track is closed, all changes are saved.

        :returns: None

        ::

           import track
           with track.load('tracks/rp_genes.bed') as t:
               t.remove('chr19_gl000209_random')
               t.save()
        """
        if self._info.modified:    self._info_write()
        if self._chrmeta.modified: self._chrmeta_write()
        self._make_missing_tables()
        self._make_missing_indexes()
        self._connection.commit()

    def _make_missing_indexes(self):
        """For every chromosomes present in the track, will create an index on the following fields if they exist:
                * start, end --> chr1_range_idx
                * score      --> chr1_score_idx
                * name       --> chr1_name_idx
        """
        if self.readonly: return
        try:
            for ch in self:
                if 'start' in self._get_fields_of_table(ch):
                    self._cursor.execute("CREATE INDEX if not exists '" + ch + "_range_idx' on '" + ch + "' (start,end)")
                if 'score' in self._get_fields_of_table(ch):
                    self._cursor.execute("CREATE INDEX if not exists '" + ch + "_score_idx' on '" + ch + "' (score)")
                if 'name' in self._get_fields_of_table(ch):
                    self._cursor.execute("CREATE INDEX if not exists '" + ch + "_name_idx' on '" +  ch + "' (name)")
        except sqlite3.OperationalError as err:
            message = "The index creation on the track '%s' failed with the following error: %s"
            raise Exception(message % (self.path, err))

    def _make_missing_tables(self):
        """Make sure every chromosome referenced in the 'chrNames' table exists as a table in the database. Will create empty tables."""
        fields = self.fields or minimum_fields
        fields = ','.join(['"' + f + '"' + ' ' + sql_field_types.get(f, 'text') for f in fields])
        for chrom_name in sorted(self.chrmeta, key=natural_sort):
            self._cursor.execute('CREATE table if not exists "' + chrom_name + '" (' + fields + ')')

    #-----------------------------------------------------------------------------#
    def rollback(self):
        """Revert all changes to the track since the last call to ``save()``.

        :returns: None

        ::

           import track
           with track.load('tracks/rp_genes.bed') as t:
               t.remove('chr19_gl000209_random')
               t.export('tmp/clean.bed')
               t.rollback()
        """
        self._connection.rollback()

    #-----------------------------------------------------------------------------#
    def vacuum(self):
        """Rebuilds the database making it shrink in file size. This method is useful when, after having executed many inserts, updates, and deletes, the SQLite file is fragmented and full of empty space.

        :returns: None

        ::

           import track
           with track.load('tracks/rp_genes.bed') as t:
               t.remove('chr19_gl000209_random')
               t.vaccum()
        """
        self._cursor.execute("VACUUM")

    #-----------------------------------------------------------------------------#
    def close(self):
        """Close the current track. This method is useful when for some special reason you are not using the ``with ... as``` form for loading tracks.

        :returns: None

        ::

            import track
            t = track.load('tracks/rp_genes.bed')
            t.remove('chr19_gl000209_random')
            t.close()
        """
        # Commit changes to the database #
        if self.modified and self.autosave: self.save()
        # Close all cursors #
        for cur in self.all_cursors: cur.close()
        # Close the connection #
        self._connection.close()
        # If the original file was not an sql #
        if not self.readonly and self.orig_path:
            # Rewrite the file #
            if os.path.exists(self.orig_path): os.remove(self.orig_path)
            convert(self.path, (self.orig_path, self.orig_format))
            # Remove the temporary SQL #
            os.remove(self.path)

    #-----------------------------------------------------------------------------#
    def export(self, path, format=None):
        """Export the current track to a given format. A new file is created at the specified path. The current track object is unchanged

        :param path: is the path to track file to create.
        :type  path: string
        :param format: is an optional parameter specifying the format of the track to create when it cannot be guessed from the file extension.
        :type  format: string
        :returns: None

        ::

            import track
            with track.load('tracks/rp_genes.bed') as t:
                t.remove('chr19_gl000209_random')
                t.export('tmp/clean.bed')
                t.rollback()
        """
        # Check it is not taken #
        check_path(path)
        # Guess format #
        if not format: format = determine_format(path)
        # Get a serializer #
        serializer = get_serializer(path, format)
        # Get a parser #
        parser = get_parser(self, 'sql')
        # Do it #
        return parser(serializer)

    #-----------------------------------------------------------------------------#
    def read(self, selection=None, fields=None, order='start,end'):
        """Read data from the track.

        :param selection: A chromosome name, or a dictionary specifying a region, see below.
        :param fields: is an optional list of fields which will influence the length of the tuples returned and the way in which the information is returned. The default is to read every field available for the given chromosome. If the *track.fields* attribute is set, that will be used.
        :type  fields: list of strings
        :param order: is am optional sublist of *fields* which will influence the order in which the tuples are yielded. By default results are sorted by ``start`` and, secondly, by ``end``.
        :type  order: list of strings

        :returns: a generator object yielding rows. A row can be referenced like a tuple or like a dictionary.

        *selection* can be the name of a chromosome, in which case all the data on that chromosome will be returned.

        *selection* can be left empty, then the data from all chromosome is returned.

        *selection* can also be a dictionary specifying: regions, score intervals or strands. If you specify a region in which case only features contained in that region will be returned. But you can also input a tuple specifying a score interval in which case only features contained in those score boundaries will be returned. You can even specify a strand. The dictionary can contain one or several of these arguments. See code example for more details.

        Adding the parameter ``'inclusion':'strict'`` to a region dictionary will return only features exactly contained inside the interval instead of features simply included in the interval. To combine multiple selections you can specify a list including chromosome names and region dictionaries. As expected, if such is the case, the joined data from those selections will be returned with an added ``chr`` field in front since the results may span several chromosomes.

        ::

            import track
            with track.load('tracks/example.sql') as t:
                data = t.read()
                data = t.read('chr2')
                data = t.read('chr3', ['name', 'strand'])
                data = t.read(['chr1','chr2','chr3'])
                data = t.read({'chr':'chr1', 'start':100})
                data = t.read({'chr':'chr1', 'start':10000, 'end':15000})
                data = t.read({'chr':'chr1', 'start':10000, 'end':15000, 'inclusion':'strict'})
                data = t.read({'chr':'chr1', 'strand':1})
                data = t.read({'chr':'chr1', 'score':(10,100)})
                data = t.read({'chr':'chr1', 'start':10000, 'end':15000, 'strand':-1, 'score':(10,100)})
                data = t.read({'chr':'chr5', 'start':0, 'end':200}, ['strand', 'start', 'score'])
        """
        # Default values #
        where = None
        ##### SELECTION #####
        if not selection: selection = self.chromosomes
        # Case list of things #
        if isinstance(selection, (list, tuple)):
            return join_read_queries(self, selection, fields)
        # Case selection dictionary #
        elif isinstance(selection, dict):
            chrom = selection['chr']
            where = " WHERE " + make_cond_from_sel(selection)
        # Case chromosome name #
        elif isinstance(selection, basestring): chrom = selection
        # Other cases #
        else: raise TypeError, 'The following selection parameter: "' + selection + '" was not understood.'
        # Empty chromosome case #
        if chrom not in self.chromosomes: return ()
        ##### FIELDS #####
        if not fields and not self._fields: query_fields = "*"
        else:
            # Columns names in the table #
            available_fields = self._get_fields_of_table(chrom)
            # Fields attribute is set or not #
            query_fields = ','.join([f in available_fields and f or py_field_types.get(f, str)().__repr__() for f in fields and fields or self._fields])
        ##### QUERY #####
        sql_command = "SELECT " + query_fields + " from '" + chrom + "'"
        # Add the where case #
        if where: sql_command += where
        # Sorting results #
        if order: sql_command += ' order by ' + order
        # Make a new cursor #
        cursor = self.cursor()
        ##### ERROR CATCHING #####
        try:
            cursor.execute(sql_command)
        except sqlite3.OperationalError as err:
            message = "The command <%s%s%s> on the track '%s' failed with error:\n\n %s%s%s"
            message = message % (Color.cyn, sql_command, Color.end, self.path, Color.u_red, err, Color.end)
            raise Exception(message)
        # Make a feature stream #
        return FeatureStream(cursor)

    #-----------------------------------------------------------------------------#
    def write(self, chromosome, data, fields=None):
        """Write data to a genomic file. Will write many feature at once into a given chromosome.

        :param chromosome: is the name of the chromosome on which one wants to write. For instance, if one is using the BED format this will become the first column, while if one is using the SQL format this will become the name of the table to be created.
        :type  chromosome: string
        :param data: must be an iterable object that yields tuples or rows of the correct length. As an example, the ``read`` function of this class produces such objects. *data* can have a *fields* attribute describing what the different elements of the tuple represent. *data* can also simply be a list of tuples.
        :type  data: an iteratable
        :param fields: is a parameter describing what the different elements in *data* represent. It is optional and is used only if *data* doesn't already have a ``fields`` attribute.
        :type  fields: list of strings

        :returns: None

        ::

            import track
            with track.load('tracks/example.sql') as t:
                t.write('chr1', [(10, 20, 'A', 0.0, 1), (40, 50, 'B', 0.0, -1)])
            with track.load('tracks/example.sql') as t:
                def example_generator():
                    for i in xrange(5):
                        yield (10, 20, 'X')
                t.write('chr2', example_generator(), fields=['start','end','name'])
            with track.load('tracks/new.sql') as t2:
                with track.load('tracks/orig.sql') as t1:
                    t1.write('chr1', t2.read('chr1'))
        """
        # Check track attributes #
        if self.readonly: return
        self._modified = True
        # Check what the data generator yields #
        if isinstance(data, FeatureStream) and data.kind == SuperRow: data.generator = imap(tuple,data)
        # Guess the fields we are getting #
        if fields:                           incoming_fields = fields
        elif hasattr(data, 'fields'):        incoming_fields = data.fields
        elif hasattr(data, 'description'):   incoming_fields = [x[0] for x in data.description]
        elif self._fields:                   incoming_fields = self._fields
        elif chromosome in self.chromosomes: incoming_fields = self._get_fields_of_table(chromosome)
        else:                                incoming_fields = default_fields
        # Current fields present in table #
        chrom_exists = chromosome in self.chromosomes
        current_fields = chrom_exists and self._get_fields_of_table(chromosome) or []
        # The fields we want to write #
        if self._fields: outgoing_fields = self._fields
        else:            outgoing_fields = incoming_fields
        # Make them sets #
        outgoing_set = set(outgoing_fields)
        incoming_set = set(incoming_fields)
        current_set  = set(current_fields)
        # Maybe we need to create the table #
        if not chrom_exists:
            fields = ','.join(['"' + field + '"' + ' ' + sql_field_types.get(field, 'text') for field in outgoing_fields])
            self._write_cursor.execute('CREATE table "' + chromosome + '" (' + fields + ')')
            current_fields = outgoing_fields
        # Or maybe we need to create new columns #
        else:
            for field in outgoing_set - current_set:
                self._write_cursor.execute('ALTER table "' + chromosome + '" ADD "' + field + '" ' + sql_field_types.get(field, 'text'))
        # Adjust size #
        if outgoing_set > incoming_set:
            outgoing_fields = incoming_fields
        if outgoing_set < incoming_set:
            indicies = tuple([incoming_fields.index(f) for f in outgoing_fields])
            data = pick_iterator_elements(data, indicies)
        # Protect names for SQL query #
        outgoing_fields = ['"' + f + '"' for f in outgoing_fields]
        question_marks = '(' + ','.join(['?' for x in xrange(len(outgoing_fields))]) + ')'
        sql_command = 'INSERT into "' + chromosome + '" (' + ','.join(outgoing_fields) + ') values ' + question_marks
        # Execute the insertion #
        try:
            self._write_cursor.executemany(sql_command, data)
        except (ValueError, sqlite3.OperationalError, sqlite3.ProgrammingError) as err:
            message1 = "The command <%s%s%s> on the track '%s' failed with error:\n %s%s%s"
            message1 = message1 % (Color.cyn, sql_command, Color.end, self.path, Color.u_red, err, Color.end)
            message2 = "\n * %sThe bindings%s: %s \n * %sYou gave%s: %s"
            message2 = message2 % (Color.b_ylw, Color.end, fields, Color.b_ylw, Color.end, data)
            message3 = "\n * %sFirst element%s: %s \n"
            message3 = message3 % (Color.b_ylw, Color.end, get_next_item(data))
            raise Exception(message1 + message2 + message3)

    #-----------------------------------------------------------------------------#
    def insert(self, chromosome, feature):
        """Insert one feature into an existing chromosome.

        :param chromosome: is the name of the chromosome into which one wants to insert.
        :type  chromosome: string
        :param feature: must be a tuple of the right size to fit into the chromosome table.
        :type  feature: tuple

        :returns: None.

        ::

            import track
            with track.load('tracks/example.sql') as t:
                t.insert('chr1', (10, 20, 'A')
        """
        question_marks = '(' + ','.join(['?' for x in xrange(len(feature))]) + ')'
        self._write_cursor.execute('insert into "' + chromosome + '" values ' + question_marks, feature)

    #-----------------------------------------------------------------------------#
    def remove(self, chromosome):
        """Remove data from a given chromosome.

        :param chromosome: is the name of the chromosome that one wishes to delete or a list of chromosomes to delete.
        :type  chromosome: string

        :returns: None.

        ::

            import track
            with track.load('tracks/example.sql') as t:
                t.remove('chr1')
            with track.load('tracks/example.sql') as t:
                t.remove(['chr1', 'chr2', 'chr3'])
        """
        # Check track attributes #
        self._modified = True
        if self.readonly: return
        # Can be a list or a string #
        if isinstance(chromosome, list):
            for x in chromosome: self.remove(x)
        else:
            self._cursor.execute("DROP table '" + chromosome + "'")
            if chromosome in self.chrmeta: self.chrmeta.pop(chromosome)

    #-----------------------------------------------------------------------------#
    def delete_fields(self, fields):
        """Remove the given fields from all chromosomes. This is equivalent to dropping full columns in the database.

        :param fields: A list of fields such as ``['score','strand']``.
        :type  fields: list

        :returns: None.

        ::

            import track
            with track.load('tracks/example.sql') as t:
                print t.fields
                t.delete_fields(['score','strand'])
                print t.fields
        """
        # Check track attributes #
        self._modified = True
        if self.readonly: return
        # SQLite doesn't support dropping columns directly #
        sql_script = '''CREATE TABLE "%(chrom)s_tmp" (%(types)s);
        INSERT INTO "%(chrom)s_tmp" SELECT %(names)s FROM "%(chrom)s";
        DROP TABLE "%(chrom)s";
        ALTER TABLE "%(chrom)s_tmp" RENAME TO "%(chrom)s";'''
        # Loop on chromosomes #
        for chrom in self:
            current_fields = self._get_fields_of_table(chrom)
            # Should anything change ? #
            if not set(fields) & set(current_fields): continue
            # What will be the fields now ? #
            new_fields = [f for f in current_fields if f not in fields]
            # Do we have anything left ? #
            if not new_fields:
                self.remove(chrom)
                return
            # Drop all the columns #
            names = ','.join(['"' + f + '"' for f in new_fields])
            types = ','.join(['"' + f + '"' + ' ' + sql_field_types.get(f, 'text') for f in new_fields])
            custom_sql_script= sql_script % dict(chrom=chrom, names=names, types=types)
            self._cursor.executescript(custom_sql_script)

    #-----------------------------------------------------------------------------#
    def rename(self, previous_name, new_name):
        """Rename a chromosome from *previous_name* to *new_name*

        :param previous_name: is the name of the chromosome that one wishes to rename.
        :type  previous_name: string
        :param new_name: is the name that that chromosome will now be referred by.
        :type  new_name: string

        :returns: None.

        ::

            import track
            with track.load('tracks/rp_genes.bed') as t:
                t.rename('chr4', 'chrIV')
        """
        # Check track attributes #
        self._modified = True
        if self.readonly: return
        # Check same name #
        if previous_name == new_name: return
        # Check previous exists #
        if previous_name not in self.chromosomes: raise Exception("The chromosome '" + previous_name + "' doesn't exist.")
        # Check new doesn't exist #
        if new_name in self.chromosomes:
            message = "The chromosome '%s' can't be renamed to '%s', as '%s' alredy exists."
            raise Exception(message % (previous_name, new_name, new_name))
        # Check different #
        if new_name == previous_name: return
        # SQL query #
        command = "ALTER TABLE '" + previous_name + "' RENAME TO '" + new_name + "'"
        try:
            self._cursor.execute(command)
        except sqlite3.OperationalError as err:
            message = "The command <%s%s%s> on the track '%s' failed with error:\n %s%s%s"
            message = message % (Color.cyn, command, Color.end, self.path, Color.u_red, err, Color.end)
            raise Exception(message)
        # Drop indexes #
        self._cursor.execute("drop index IF EXISTS '" + previous_name + "_range_idx'")
        self._cursor.execute("drop index IF EXISTS '" + previous_name + "_score_idx'")
        self._cursor.execute("drop index IF EXISTS '" + previous_name + "_name_idx'")
        # Rename the chrmeta #
        if previous_name in self.chrmeta:
            self.chrmeta[new_name] = self.chrmeta[previous_name]
            self.chrmeta.pop(previous_name)

    #-----------------------------------------------------------------------------#
    def search(self, query_dict, fields=None, chromosome=None, exact_match=False):
        """Search for parameters inside your track. You can specify several parameters.

        :param selection: list of the fields you want to have in the result (to insure that all result will have the same number of columns)
        :param query_dict: A dictionary specifying keys and values to search for. See examples.
        :type  query_dict: dict
        :param chromosome: Optionally, the name of the chromosome on which one wants to search. If ``None``, the search is performed on all chromosomes and every feature contains a new field specifying its chromosome.
        :type  chromosome: string
        :param exact_match: By default, will find all entries which contain the query. If set to ``True``, will only find entries that exactly match the query.
        :type  exact_match: bool

        :returns: a generator object yielding rows. A row can be referenced like a tuple or like a dictionary.

        ::

           import track
           with track.load('tracks/rp_genes.bed') as t:
               results = t.search({'gene_id':3})
               results = t.search({'gene_id':3, 'gene_name':'YCCA3'}, 'chr1')
        """
        # A example final SQL string generated by this function is:
        # SELECT * from ("chrI","chrII","chrIII") WHERE gene_id like '%3%' AND gene_name like '%YCCA3%'
        conditions = []
        if fields is None :
            fields = self.fields
        # Generate condition #
        for k,v in query_dict.items():
            if exact_match: conditions.append(' %s = "%s" ' % (k,v))
            else:           conditions.append(' %s like "%%%s%%" ' % (k,v))
        where = ' WHERE ' + ' AND '.join(conditions)
        # Iterate on chromosomes #
        if chromosome: query_str = 'SELECT ' + ', '.join(fields) + ' from "%s"' % chromosome + where
        else:          query_str = ' UNION '.join(['SELECT "%s",' % chrom + ', '.join(fields) + ' from "%s"' % chrom + where for chrom in self])
        # Execute it #
        cursor = self.cursor()
        return cursor.execute(query_str)

    #-----------------------------------------------------------------------------#
    def find_column_name(self, synonyms, table_name=None):
        """Search for a field name in the database. This can be useful when a specific field name has several synonyms and you don't know which one your data source is using. If several names match, will only return the first match.

        :param synonyms: Several names commonly used for the same thing.
        :type  synonyms: list
        :param table_name: Optionally, a specific chromosome to search on.
        :type  table_name: list

        :returns: An element of **synonyms** or an empty string if no match is found.

        ::
            with track.load('tracks/rp_genes.bed') as t:
                name = t.find_column_name(['name', 'gene_name', 'gname', 'gene name'])
                > gene_name
        """
        if table_name is not None: fields = self.fields
        else:                      fields = self._get_fields_of_table(table_name)
        for name in synonyms:
            if name in self.fields: return name
        return ''

    #-----------------------------------------------------------------------------#
    def count(self, selection=None):
        """Count the number of features or entries in a given selection.

        :param selection: is the name of a chromosome, a list of chromosomes, a particular span or a list of spans. In other words, a value similar to the *selection* parameter of the *read* method. If left empty, will count every feature in a track

        :returns: an integer.

        ::

            import track
            with track.load('tracks/example.sql') as t:
                num = t.count('chr1')
                num = t.count(['chr1','chr2','chr3'])
                num = t.count({'chr':'chr1', 'start':10000, 'end':15000})
        """
        # Default selection #
        if not selection:
            selection = self.chromosomes
        # Case several chromosome #
        if isinstance(selection, list) or isinstance(selection, tuple):
            return sum([self.count(s) for s in selection])
        # Case chromosome name #
        elif isinstance(selection, basestring):
            if selection not in self.chromosomes: return 0
            sql_request = "select COUNT(*) from '" + selection + "'"
        # Case span dictionary #
        elif isinstance(selection, dict):
            chrom = selection['chr']
            if chrom not in self.chromosomes: return 0
            sql_request = "select COUNT(*) from '" + chrom + "' where " + make_cond_from_sel(selection)
        # Other cases #
        else: raise TypeError, 'The following selection parameter: "' + selection + '" was not understood'
        # Return the results #
        cursor = self.cursor()
        return cursor.execute(sql_request).fetchone()[0]

    #-----------------------------------------------------------------------------#
    def ucsc_to_ensembl(self):
        """Convert all entries of a track from the UCSC standard to the Ensembl standard effectively adding one to every start position.

        :returns: None.

        ::

           import track
           with track.load('tracks/example.sql') as t:
               t.ucsc_to_ensembl()
        """
        for chrom in self.chromosomes: self._cursor.execute("update '" + chrom + "' set start=start+1")

    def ensembl_to_ucsc(self):
        """Converts all entries of a track from the Ensembl standard to the UCSC standard effectively subtracting one from every start position.

        :returns: None.

        ::

           import track
           with track.load('tracks/rp_genes.bed') as t:
               t.ensembl_to_ucsc()
        """
        for chrom in self.chromosomes: self._cursor.execute("update '" + chrom + "' set start=start-1")

    #-----------------------------------------------------------------------------#
    def get_full_score_vector(self, chromosome):
        """Create an iterable with as many elements as there are base pairs in the chromosomes specified by the *chromosome* parameter. Every element of the iterable is a float indicating the score at that position. If the track has no score associated, ones are inserted where features are present.

        :param chromosome: is the name of the chromosome on which one wants to create a score vector from.
        :type  chromosome: string

        :returns: an iterable yielding floats.

        ::

            import track
            with track.new('tmp/track.sql') as t:
                scores = t.get_full_score_vector('chr1')
        """
        # Check chromosome existence #
        if chromosome not in self: return
        chr_length = self.chrmeta[chromosome]['length'] if chromosome in self.chrmeta else None
        # Call read #
        fields = ['start','end','score'] if 'score' in self.fields else ['start','end']
        data = self.read(chromosome, fields)
        # Special function for tracks without score #
        add_ones = lambda X: (tuple(x) + (1.0,) for x in X)
        if 'score' not in self.fields: data = add_ones(data)
        # Initialization #
        last_end = 0
        x = (-1,0)
        # Core loop #
        for x in data:
            for i in xrange(last_end, x[0]): yield 0.0
            for i in xrange(x[0],     x[1]): yield x[2]
            last_end = x[1]
        # End piece #
        if chr_length:
            for i in xrange(x[1], chr_length): yield 0.0

    #-----------------------------------------------------------------------------#
    def get_partial_score_vector(self, chromosome, start, end):
        """Create an iterable with as many elements as there are base pairs in the interval between *start* and *end*. Every element of the iterable is a float indicating the score at that position. If the track has no score associated, ones are inserted where features are present.

        :param chromosome: is the name of the chromosome on which one wants to create a score vector from.
        :type  chromosome: string
        :param start: The base pair position where scores will start being read from. Defaults to 0.
        :type  start: int
        :param end: The base pair position where scores will stop being read from. Defaults to the length of the chromosome.
        :type  end: int

        :returns: an iterable yielding floats.

        ::

            import track
            with track.new('tmp/track.sql') as t:
                scores = t.get_score_vector('chr1')
        """
        # Check chromosome existence #
        if chromosome not in self: return
        # Call read #
        selection = {'chr': chromosome, 'start': start, 'end': end}
        fields = ['start','end','score'] if 'score' in self.fields else ['start','end']
        data = self.read(selection, fields)
        # Special function for tracks without score #
        add_ones = lambda X: (tuple(x) + (1.0,) for x in X)
        if 'score' not in self.fields: data = add_ones(data)
        # Core loop #
        x = (start, start)
        for x in data:
            if start >= x[1]: continue
            if start < x[0]:
                for i in xrange(start, x[0]): yield 0.0
                start = x[0]
            if end <= x[1]:
                for i in xrange(start, end): yield x[2]
                break
            else:
                for i in xrange(start,  x[1]): yield x[2]
                start = x[1]
        # End piece #
        for i in xrange(x[1], end): yield 0.0

    #-----------------------------------------------------------------------------#
    def roman_to_integer(self, names=None):
        """Convert the name of all chromosomes from the roman numeral standard to the arabic numeral standard. For instance, 'chrI' will become 'chr1' while 'chrII' will become 'chr2', etc.

        :param names: an optional dictionary specifying how to translate particular cases. Example: ``{'chrM':'chrQ', '2micron':'chrR'}``
        :type  names: dict

        :returns: None.

        ::

            import track
            with track.new('tmp/track.sql') as t:
                scores = t.roman_to_integer()
        """
        names = names or {'chrM':'chrQ', '2micron':'chrR'}
        def convert(chrom):
            if chrom in names: return names[chrom]
            match = re.search('([a-zA-Z]*?)([IVX]+)$', chrom)
            if match: return match.group(1) + str(roman_to_int(match.group(2)))
            else: return chrom
        for chrom in self: self.rename(chrom, convert(chrom))

    def integer_to_roman(self, names=None):
        """Convert the name of all chromosomes from the arabic numeral standard to the roman numeral standard. For instance, 'chr1' will become 'chrI' while 'chr2' will become 'chrII', etc.

        :param names: an optional dictionary specifying how to translate particular cases. Example: ``{'chrQ':'chrM', 'chrR':'2micron'}``
        :type  names: dict

        :returns: None.

        ::

            import track
            with track.new('tmp/track.sql') as t:
                scores = t.roman_to_integer()
        """
        names = names or {'chrQ':'chrM', 'chrR':'2micron'}
        def convert(chrom):
            if chrom in names: return names[chrom]
            match = re.search('([a-zA-Z]*)([0-9]+)$', chrom)
            if match: return match.group(1) + int_to_roman(int(match.group(2)))
            else: return chrom
        for chrom in self: self.rename(chrom, convert(chrom))

    #--------------------------------------------------------------------------#
    @property
    def info(self):
        """A dictionary of meta data associated to the track (information like the source, etc). For instance:

              ``{'datatype': 'signal', 'source': 'SGD', 'orig_name': 'splice_sites.bed'}``
        """
        return self._info

    @info.setter
    def info(self, value):
        self._info.overwrite(value)

    def _info_read(self):
        """Populate the *self.info* attribute with information found in the 'attributes' table."""
        if not 'attributes' in self.tables: return
        # Make a dictionary directly from the table #
        query = self._cursor.execute('SELECT key, value from "attributes"')
        self.info = dict(query.fetchall())
        # Freshly loaded, so not modified #
        self.info.modified = False

    def _info_write(self):
        """Rewrite the 'attributes' table so that it reflects the contents of the *self.info* attribute."""
        if self.readonly: return
        self._cursor.execute('DROP table IF EXISTS "attributes"')
        if not self.info: return
        # Write every dictionary entry #
        self._cursor.execute('CREATE table "attributes" ("key" text, "value" text)')
        for k in sorted(self.info.keys(), key=natural_sort):
            self._cursor.execute('INSERT into "attributes" ("key","value") values (?,?)', (k, self.info[k]))

    @property
    def datatype(self):
        """Giving a datatype to your track is optional. The default datatype is ``None``. Other possible datatypes are ``features``, ``signal`` or ``relational``. Changing the datatype imposes some conditions on the entries that the track contains. This attribute is stored inside the *info* dictionary.

        ::

            import track
            with track.new('tmp/track.sql') as t:
                t.datatype = 'signal'
        """
        return self.info.get('datatype', None)

    @datatype.setter
    def datatype(self, value):
        if value not in ['features', 'signal', 'relational']:
            raise Exception("The datatype you are trying to use is invalid: '" + str(value) + "'.")
        self.info['datatype'] = value

    @property
    def name(self):
        """Giving a name to your track is optional. The default name is the filename. This attribute is stored inside the *info* dictionary."""
        return self.info.get('name', os.path.basename(self.path))

    @name.setter
    def name(self, value):
        self.info['name'] = value

    #-----------------------------------------------------------------------------#
    @property
    def chrmeta(self):
        """Contains extra chromosomal meta data such as chromosome length information. *chrmeta* is a dictionary where each key is a chromosome names. For instance:

             ``{'chr1': {'length': 197195432}, 'chr2': {'length': 129993255}}``

        You would hence use it like this::

            import track
            with track.load('tmp/track.sql') as t:
                print t.chrmeta['chr1']['length']

        Of course, genomic formats such as ``bed`` cannot store this kind of meta data. Hence, when loading tracks in these text formats, this information is lost once the track is closed."""
        return self._chrmeta

    @chrmeta.setter
    def chrmeta(self, value):
        self._chrmeta.overwrite(value)

    def _chrmeta_read(self):
        """Populate the self.chrmeta attribute with information found in the 'chrNames' table."""
        # If the table doesn't exist, just use the names
        if not 'chrNames' in self.tables:
            dictionary = dict([(chrom, dict()) for chrom in self])
        else:
            # Columns are the chromosome attributes #
            # ['name', 'length']
            columns = self._get_fields_of_table("chrNames")
            # Rows are the chromosome names #
            # [{'name': 'chr1', 'length': 1000}, {'name': 'chr2', 'length': 2000}]
            query = self._cursor.execute('SELECT * from "chrNames"').fetchall()
            rows = [dict([(k,r[i]) for i, k in enumerate(columns)]) for r in query]
            # Make a pretty dictionary of dictionaries #
            # {'chr1': {'length': 1000}, 'chr2': {'length': 2000}}
            dictionary = dict([(r['name'], dict([(k, r[k]) for k in columns if k != 'name'])) for r in rows])
        # Freshly loaded, so not modified #
        self.chrmeta = dictionary
        self.chrmeta.modified = False

    def _chrmeta_write(self):
        """Rewrite the 'chrNames' table so that it reflects the contents of the self.chrmeta attribute."""
        if self.readonly: return
        self._cursor.execute('DROP table IF EXISTS "chrNames"')
        if not self.chrmeta: return
        # Rows are the chromosome names #
        # [{'name': 'chr1', 'length': 1000}, {'name': 'chr2', 'length': 2000}]
        rows = [dict([('name', chrom)] + [(k,v) for k,v in self.chrmeta[chrom].items()]) for chrom in self.chrmeta]
        # Columns are the field names #
        # ['name', 'length']
        columns = rows[0].keys()
        fields = ','.join(['"' + field + '"' + ' ' + sql_field_types.get(field, 'text') for field in columns])
        self._cursor.execute('CREATE table "chrNames" (' + fields + ')')
        for r in sorted(rows, key=lambda x: natural_sort(x['name'])):
            question_marks = '(' + ','.join(['?' for x in r.keys()]) + ')'
            column_names   = '(' + ','.join(['"' + k + '"' for k in r.keys()]) + ')'
            cell_values    = tuple(r.values())
            self._cursor.execute('INSERT into "chrNames" ' + column_names + ' values ' + question_marks, cell_values)

    def load_chr_file(self, path):
        """Set the *chrmeta* attribute of the track by loading a chromosome file. The chromosome file is structured as tab-separated text file containing two columns: the first specifies a chromosomes name and the second its length as an integer.

        :param path: is the file path to the chromosome file to load.
        :type  path: string

        :returns: None.
        """
        self.chrmeta = parse_chr_file(path)

    def export_chr_file(self, path):
        """Output the information contained in the *chrmeta* attribute into a plain text file. The chromosome file is structured as tab-separated text file containing two columns: the first specifies a chromosomes name and the second its length as an integer

        :param path: is the file path to the chromosome file to create.
        :type  path: string

        :returns: None.
        """
        serialize_chr_file(self.chrmeta, path)

    @property
    def assembly(self):
        """Giving an assembly to your track is optional. However, if you set this variable for your track, you should input a valid assembly name such as 'sacCer2'. Doing so will set the *chrmeta* attribute and rename all the chromosome to their canonical names if a correspondence is found. You can also call ``guess_assembly()``. This attribute is also stored inside the *info* dictionary.

        ::

            import track
            track.convert('tracks/genes.bed', 'tracks/genes.sql')
            with track.load('tracks/genes.sql') as t:
                t.assembly = 'hg19'
        """
        return self.info.get('assembly', 'Unnamed')

    @assembly.setter
    def assembly(self, value):
        # Get the assembly #
        assembly = Assembly(value)
        # Add the attribute #
        self.info['assembly'] = value
        # Check if the tables need renaming or deleting #
        for orig_name in self.chromosomes:
            cannonical_name = assembly.guess_chromosome_name(orig_name)
            if cannonical_name: self.rename(orig_name, cannonical_name)
            else: self.remove(orig_name)
        # Add the chrmeta #
        self.chrmeta = assembly.chrmeta

    #-----------------------------------------------------------------------------#
    def aggregated_read(self, selection, base_fields, order_by=None):
        """
        Read the track with the *base_fields*, and aggregate all supplementary fields
        found in a single string.

        :param selection: The **selection** argument is passed to the **read** function.
        :type selection: list
        :param base_fields: the columns you don't want to aggregate.
        :type base_fields: list
        :param order_by: The **order_by** argument is passed to the **read** function.
        :type order_by: list

        :returns: a generator object yielding rows.

        ::

            >> import track
            >> t = track.load('tmp/track.sql')
            >> t.fields
            ['start', 'stop', 'score', 'name', 'type', 'strand']
            >> for feature in t.read('chr1'): print feature
            [1,  10, 2.2, 'tc36', 'transcript', '+']
            [1,  10, 3.4, 'tc36', 'transcript', '-']
            [11, 12, 3.4, 'tc37', 'transcript', '+']
            >> data = t.aggregated_read('chr1', ('start', 'stop', 'score'))
            >> for feature in data: print feature
            [1,  10, 2.2, "{name :'tc36', type:'transcript', strand:'+'}"]
            [1,  10, 3.4, "{name :'tc36', type:'transcript', strand:'-'}"]
            [11, 12, 3.4, "{name :'tc37', type:'transcript', strand:'+'}"]
        """
        base_fields = tuple(base_fields)
        track_fields = self.fields
        supplementary_fields = tuple(set(track_fields) - set(base_fields))
        blen = len(base_fields)
        ll = range(blen)
        for feature in self.read(selection, base_fields + supplementary_fields, order=order_by):
            d = [{field : feature[field]} for field in supplementary_fields if field in feature.keys()]
            yield [feature[i] for i in ll] + [json.dumps(d)]

################################################################################
class FeatureStream(object):
    """Contains an iterator yielding features and an extra
       fields attribute.

       @param data: the iterator (or cursor) itself.
       @param fields: the list of fields
    """

    def __init__(self, generator, fields=None, kind=None):
        # The generator itself #
        self.generator = generator
        # The type of elements yielded #
        if not kind and hasattr(generator, 'connection'): kind = generator.connection.row_factory or tuple
        self.kind = kind
        # The description of the elements inside #
        if not fields and hasattr(generator, 'description'): fields = [x[0] for x in generator.description]
        self.fields = fields

    def __repr__(self): return "FeatureStream containing %s" % self.generator

    def __iter__(self): return self.generator

    def next(self): return self.generator.next()

#-----------------------------------#
# This code was written by the BBCF #
# http://bbcf.epfl.ch/              #
# webmaster.bbcf@epfl.ch            #
#-----------------------------------#



"""
Contains chained write tests.
"""

# Built-in modules #
import os

# Internal modules #
import track
from track.test import samples
from track.common import temporary_path

# Unittesting module #
try:
    import unittest2 as unittest
except ImportError:
    import unittest

# Nosetest flag #
__test__ = True

###################################################################################
class TestChain(unittest.TestCase):
    """Chain a read to a write"""
    def runTest(self):
        in_path = samples['small_signals'][7]['sql']
        out_path = temporary_path('.sql')
        with track.load(in_path) as i:
            with track.new(out_path) as o:
                for chrom in i: o.write(chrom, i.read(chrom))
                self.assertEqual(list(o.read('chrI')), list(i.read('chrI')))
        os.remove(out_path)

#---------------------------------------------------------------------------------#
class TestMissingFields(unittest.TestCase):
    """Write with missing fields"""
    def runTest(self):
        in_path = samples['small_features'][1]['sql']
        out_path = temporary_path('.sql')
        chrom = 'chrI'
        with track.load(in_path) as i:
            with track.new(out_path) as o:
                o.fields = track.default_fields
                o.write(chrom, i.read(chrom, ('start','end')))
                got = tuple(o.read(chrom).next())
                expected = (0, 10, None, None, None)
        self.assertEqual(got, expected)
        os.remove(out_path)

#-----------------------------------#
# This code was written by the BBCF #
# http://bbcf.epfl.ch/              #
# webmaster.bbcf@epfl.ch            #
#-----------------------------------#



"""
This module implements the parsing of the in-memory data.
In this case, self.path will be a large dictionary.
"""

# Internal modules #
from track.parse import Parser

################################################################################
class ParserMemory(Parser):
    format = 'memory'
    def parse(self):
        self.handler.newTrack({}, self.name)
        for chrom in self.path:
            for feature in self.path[chrom]:
                self.handler.newFeature((chrom,) + feature)

#-----------------------------------#
# This code was written by the BBCF #
# http://bbcf.epfl.ch/              #
# webmaster.bbcf@epfl.ch            #
#-----------------------------------#



# Built-in modules #
import os

# Internal modules #
from track.util import determine_format
from track.common import check_path
from track.parse import get_parser
from track.serialize import get_serializer

################################################################################
def read(path, format=None):
    """Loads a track from disk, whatever the format is and puts it in an enourmous dictionary. Contrary to most of the algorithms in this package, this method will load everything into the RAM. Be careful when using this method on large tracks.

       :param path: is the path to track file to create.
       :type  path: string
       :param format: is an optional parameter specifying the format of the track to create when it cannot be guessed from the file extension.
       :type  format: string
       :returns: a dictionary containing all features of every chromosome contained in a track.

       ::

            from track.memory import read, write
            genes = read('tracks/rp_genes.bed')
            print genes['chrX'][1523]
    """
    # Guess the format #
    if not format: format = determine_format(path)
    # Do the job #
    serializer = get_serializer(None, 'memory')
    parser     = get_parser(path, format)
    return parser(serializer)

################################################################################
def write(dictionary, path, format=None):
    """Write a dictionary to disk, whatever the format is.

       :param dictionary: contains the chromosomes and features.
       :type  dictionary: dict
       :param format: is the path to track file to write to.
       :type  format: string
       :param format: is an optional parameter specifying the format of the track to create when it cannot be guessed from the file extension.
       :type  format: string
       :returns: None

       ::

            from track.memory import read, write
            genes = read('tracks/rp_genes.gff')
            genes['chrX'][1523] = (183, 195, 'bra', 1.9)
            write('tracks/modified.bed', genes)
    """
    # Guess the format #
    if not format: format = os.path.splitext(path)[1][1:]
    check_path(path)
    # Do the job #
    serializer = get_serializer(path, format)
    parser     = get_parser(dictionary, 'memory')
    return parser(serializer)

#-----------------------------------#
# This code was written by the BBCF #
# http://bbcf.epfl.ch/              #
# webmaster.bbcf@epfl.ch            #
#-----------------------------------#



"""
This module implements the parsing of BED files according to this standard:

http://genome.ucsc.edu/FAQ/FAQformat.html#format3
"""

# Built-in modules #
import shlex

# Internal modules #
from track.parse import Parser
from track.common import iterate_lines
from track.util import strand_to_int

# Constants #
all_fields = ['start', 'end', 'name', 'score', 'strand', 'thick_start',
              'thick_end', 'item_rgb', 'block_count', 'block_sizes', 'block_starts']

################################################################################
class ParserBED(Parser):
    format = 'bed'
    def parse(self):
        # Initial variables #
        fields = []
        info   = {}
        # Main loop #
        for number, line in iterate_lines(self.path):
            # Ignored lines #
            if line.startswith("browser "): continue
            # Track headers #
            if line.startswith("track "):
                try:
                    info = dict([p.split('=',1) for p in shlex.split(line[6:])])
                except ValueError:
                    self.handler.error("The track%s seems to have an invalid <track> header line", self.path, number)
                fields = []
                continue
            # Split the lines #
            items = line.split('\t')
            if len(items) == 1: items = line.split()
            # Chromosome #
            chrom = items.pop(0)
            # Have we started a track already ? #
            if not fields:
                self.handler.newTrack(info, self.name)
                fields = all_fields[0:len(items)]
                self.handler.defineFields(fields)
            # Start and end fields #
            try:
                items[0] = int(items[0])
                items[1] = int(items[1])
            except ValueError:
                self.handler.error("The track%s has non integers as interval bounds", self.path, number)
            except IndexError:
                self.handler.error("The track%s has less than two columns", self.path, number)
            # All following fields are optional #
            try:
                # Name field #
                if items[2] == '.': items[2] = ''
                # Score field #
                if items[3] == '.' or items[3] == '': items[3] = 0.0
                try:
                    items[3] = float(items[3])
                except ValueError:
                    self.handler.error("The track%s has non floats as score values", self.path, number)
                # Strand field #
                items[4] = strand_to_int(items[4])
                # Thick starts #
                try:
                    items[5] = float(items[5])
                except ValueError:
                    self.handler.error("The track%s has non integers as thick starts", self.path, number)
                # Thick ends #
                try:
                    items[6] = float(items[6])
                except ValueError:
                    self.handler.error("The track%s has non integers as thick ends", self.path, number)
                # Too many fields #
                if len(items) > 11:
                    self.handler.error("The track%s has more than twelve columns", self.path, number)
            # All index errors are ignored since the fields above three are optional #
            except IndexError:
                pass
            finally:
                self.handler.newFeature(chrom, items)

#-----------------------------------#
# This code was written by the BBCF #
# http://bbcf.epfl.ch/              #
# webmaster.bbcf@epfl.ch            #
#-----------------------------------#



"""
Contains tests for the track.search method.
"""

# Internal modules #
import track
from track.test import samples

# Unittesting module #
try:
    import unittest2 as unittest
except ImportError:
    import unittest

# Nosetest flag #
__test__ = True

###################################################################################
class TestTwoParameters(unittest.TestCase):
    """A simple case with two fields"""
    def runTest(self):
        in_path = samples['small_features'][1]['sql']
        with track.load(in_path) as t:
            data = t.search({'start':'2','name':'feature'})
            got = list([tuple(x) for x in data])
        expected = [(u'chrI', 2, 8, u'Validation feature 2', 0.0),
                    (u'chrI', 20, 30, u'Validation feature 3', 10.0),
                    (u'chrI', 25, 30, u'Validation feature 4', 0.0),
                    (u'chrI', 120, 130, u'Validation feature 11', 10.0),
                    (u'chrI', 125, 135, u'Validation feature 12', 5.0)]
        self.assertEqual(got, expected)

class TestNoResults(unittest.TestCase):
    """A case with no results"""
    def runTest(self):
        in_path = samples['small_features'][1]['sql']
        with track.load(in_path) as t:
            data = t.search({'start':'2','name':'feature'}, exact_match=True)
            got = list(data)
        expected = []
        self.assertEqual(got, expected)

#-----------------------------------#
# This code was written by the BBCF #
# http://bbcf.epfl.ch/              #
# webmaster.bbcf@epfl.ch            #
#-----------------------------------#



"""
This module implements the bedgraph serialization.
"""

# Internal modules #
from track.serialize import Serializer
from track.common import format_float
from track.parse.bigwig import bigwig_to_bedgraph

# Constants #
all_fields = ['start', 'end', 'score']

################################################################################
class SerializerBedgraph(Serializer):
    format = 'bedgraph'

    def __enter__(self):
        # Special case #
        if self.parser.format == 'bigWig': return self
        # Open file #
        else: self.file = open(self.path, 'w')
        # Must return self #
        return self

    def __exit__(self, errtype, value, traceback):
        # Special case #
        if self.parser.format == 'bigWig': bigwig_to_bedgraph(self.parser.path, self.path)
        # Close file #
        else: self.file.close()

    def defineFields(self, fields):
        self.indices = []
        for f in all_fields:
            try:
                self.indices.append(fields.index(f))
            except ValueError:
                message = "You tried to write a bedgraph file without a '%s' field. Required fields are: %s"
                self.error(message % (f, all_fields))

    def newTrack(self, info=None, name=None):
        if not info: info = {}
        info['type'] = 'bedGraph'
        info['converted_by'] = __package__
        self.file.write("track " + ' '.join([k + '="' + v + '"' for k, v in info.items()]) + '\n')
        self.tracks.append(self.path)

    def newFeature(self, chrom, feature):
        # Put the fields in the right order #
        line = [feature[i] for i in self.indices]
        # Convert the score #
        line[2] = format_float(line[2])
        # Make sure eveything is a string #
        line = [str(f) for f in line]
        # Write one line #
        self.file.write(' '.join([chrom] + line) + '\n')

#-----------------------------------#
# This code was written by the BBCF #
# http://bbcf.epfl.ch/              #
# webmaster.bbcf@epfl.ch            #
#-----------------------------------#



"""
This module implements the bigwig serialization.

It requires the binary executable "bedGraphToBigWig" to be installed and listed in the $PATH.
This executable is available here:

http://hgdownload.cse.ucsc.edu/admin/exe/
"""

# Built-in modules #
import os

# Internal modules #
from track.serialize.bedgraph import SerializerBedgraph
from track.common import temporary_path, check_executable, run_tool
from track.util import serialize_chr_file

# Constants #
all_fields = ['start', 'end', 'score']

################################################################################
class SerializerBigwig(SerializerBedgraph):
    def __enter__(self):
        # Just serialize it as a bedgraph first #
        self.tmp_path = temporary_path('.bedgraph')
        self.file = open(self.tmp_path, 'w')
        return self

    def __exit__(self, errtype, value, traceback):
        self.file.close()
        bedgraph_to_bigwig(self.tmp_path, self.chrmeta, self.path)
        self.tracks = [self.path]
        os.remove(self.tmp_path)

################################################################################
def bedgraph_to_bigwig(bedgraph_path, chrmeta, bigwig_path):
    """
    Converts a bedgraph file to a bigwig file.

    :param bedgraph_path: The path to the bedgraph file to read.
    :type  bedgraph_path: string
    :param chrmeta: The chromosome metadata.
    :type  chrmeta: dict
    :param bigwig_path: The path to the bedbig file to create.
    :type  bigwig_path: string
    :returns: None
    """
    # Check the binary tool exists #
    check_executable('bedGraphToBigWig')
    # Make the chr file #
    chrfile_path = temporary_path('.chr')
    serialize_chr_file(chrmeta, chrfile_path)
    # Run the tool #
    run_tool('bedGraphToBigWig', [bedgraph_path, chrfile_path, bigwig_path])
    # Remove the chr file #
    os.remove(chrfile_path)

#-----------------------------------#
# This code was written by the BBCF #
# http://bbcf.epfl.ch/              #
# webmaster.bbcf@epfl.ch            #
#-----------------------------------#



"""
Contains tests for the SGA format.
"""

# Built-in modules #
import os

# Internal modules #
import track
from track.common import temporary_path, assert_file_equal
from track.test import samples

# Unittesting module #
try:
    import unittest2 as unittest
except ImportError:
    import unittest

# Nosetest flag #
__test__ = True

###################################################################################
class TestImport(unittest.TestCase):
    def runTest(self):
        for num, info in sorted(samples['sga_tracks'].items()):
            # Prepare paths #
            orig_sga_path = info['sga']
            orig_sql_path = info['sql']
            test_sql_path = temporary_path('.sql')
            # From SGA to SQL #
            track.convert(orig_sga_path, test_sql_path, assembly='hg19')
            self.assertTrue(assert_file_equal(orig_sql_path, test_sql_path))
            # Clean up #
            os.remove(test_sql_path)

class TestExport(unittest.TestCase):
    def runTest(self):
        info = samples['small_signals'][1]
        # Prepare paths #
        orig_sql_path = info['sql']
        orig_sga_path = info['sga']
        test_sga_path = temporary_path('.sga')
        # From SGA to SQL #
        track.convert(orig_sql_path, test_sga_path, assembly='hg19')
        self.assertTrue(assert_file_equal(orig_sga_path, test_sga_path))
        # Clean up #
        os.remove(test_sga_path)

#-----------------------------------#
# This code was written by the BBCF #
# http://bbcf.epfl.ch/              #
# webmaster.bbcf@epfl.ch            #
#-----------------------------------#



"""
This module implements the parsing of bigwig files.

http://genome.ucsc.edu/goldenPath/help/bigWig.html

It requires the binary executable "bigWigToBedGraph" to be installed and listed in the $PATH.
This executable is available here:

http://hgdownload.cse.ucsc.edu/admin/exe/

I started the implementation thinking that I would use the wigToBigWig and bigWigToWig tools provided by UCSC to come and go from the binary format. However, looking more closely at the effects of both programs, it appears that bigWigToWig does not perform as advertised and doesn't create wig files. In reality, the bigWigToWig binary takes bigwig files as input and produces bedgraph files as output. Indeed, the bigWigToWig executable has suspiciously the exact same number of bytes as the bigWigToBedGraph executable...
"""

# Built-in modules #
import os

# Internal modules #
import track
from track.parse import Parser
from track.common import temporary_path, check_executable, run_tool

################################################################################
class ParserBigwig(Parser):
    format = 'bigwig'
    def parse(self):
        # Special case #
        if self.handler.format == 'bedGraph': return
        # The two paths #
        bigwig_path   = self.path
        bedgraph_path = temporary_path('.bedgraph')
        # Run the tool #
        bigwig_to_bedgraph(bigwig_path, bedgraph_path)
        # Now just parse the bedgraph #
        parser = track.parse.get_parser(bedgraph_path, 'bedgraph')
        parser.name = self.name
        parser(self.handler)
        # Erase the temporary file #
        os.remove(bedgraph_path)

################################################################################
def bigwig_to_bedgraph(bigwig_path, bedgraph_path):
    """
    Converts a bedgraph file to a bigwig file.

    :param bigwig_path: The path to the bedbig file to read.
    :type  bigwig_path: string
    :param bedgraph_path: The path to the bedgraph file to create.
    :type  bedgraph_path: string
    :returns: None
    """
    # Check the binary tool exists #
    check_executable('bigWigToBedGraph')
    # Run the tool #
    run_tool('bigWigToBedGraph', [bigwig_path, bedgraph_path])

#-----------------------------------#
# This code was written by the BBCF #
# http://bbcf.epfl.ch/              #
# webmaster.bbcf@epfl.ch            #
#-----------------------------------#



# coding: utf-8

################################### Modules ####################################
import sys
import types
from track import common

################################# Description ##################################
label           = 'custom_boolean'
short_name      = 'Custom boolean'
long_name       = 'Execute a custom boolean'
input_tracks    = [{'key':'n_tracks', 'position':1, 'kind':'many',
                    'fields':['start','end','name','score','strand']}]
input_args      = [{'key':'fn', 'position':2, 'type':types.FunctionType,
                    'doc':'A function that takes a vector and returns a boolean.'},
                   {'key':'win_size', 'position':3, 'type': int, 'default': 1000,
                    'doc':'The number of basepairs loaded in memory at each iteration (default: 1000).'}]
input_meta      = []
output_tracks   = [{'position':1, 'fields': ['start','end']}]
tracks_collapse = None
fields_collapse = None
chroms_collapse = 'union'

################################ Documentation #################################
tooltip = \
"""
The ``custom_boolean`` manipulation takes several tracks *t_1*,..,*t_n* and a
boolean function *fn* as input, and returns one track that is *fn(t_1,..,t_n)*.
The window size *win_size* determines how many basepairs of each track are
loaded in memory at a time (default: 1000).
"""

tickets = []

################################## Examples ####################################
numeric_example = \
"""
X1 [start,end]: [(4,5),(7,9),(10,12)]
X2 [start,end]: [(1,3),(4,5),(11,14)]
X3 [start,end]: [(9,13)]
fn: not(X1) and (X2 or X3)
R               [(1,3),(9,10),(12,14)]
"""

visual_example = \
u"""
X1: ───▤▤▤▤▤▤▤▤─────────────────────
X2: ─▤▤▤▤▤▤─────────────────────────
X3: ─────▤▤▤▤▤▤▤▤▤──────────────────
fn: X1 and X2 and X3
R:  ─────▤▤─────────────────────────
"""

#################################### Tests #####################################
def test_bool_fn(b):
    return not(b[0]) and (b[1] or b[2])

def fill(l):
    return [(x[0],x[1],'',0.0,0) for x in l]

tests = [
    # Test AND
    {'tracks':   {'n_tracks': [iter(fill([(4,5),(7,9),(8,9),(10,12),(14,16),(21,28)])),
                               iter(fill([(1,3),(4,5),(11,14),(15,18),(19,20)])),
                               iter(fill([(9,13),(14,19),(22,25),(32,34)]))] },
     'args':     {'fn': lambda x: all(x), 'win_size':10},
     'chromosome': 'chrI',
     'expected': fill([(11,12),(15,16)])},

    # Test OR
    {'tracks':   {'n_tracks': [iter(fill([(4,5),(7,9),(10,12)])),
                               iter(fill([(1,3),(4,5),(11,14)])),
                               iter(fill([(9,13)]))] },
     'args':     {'fn': lambda x: any(x)},
     'chromosome': 'chrI',
     'expected': fill([(1,3),(4,5),(7,14)])},

    # Test custom fct
    {'tracks':   {'n_tracks': [iter(fill([(4,5),(7,9),(10,12)])),
                               iter(fill([(1,3),(4,5),(11,14)])),
                               iter(fill([(9,13)]))] },
     'args':     {'fn': test_bool_fn, 'win_size': 5},
     'chromosome': 'chrI',
     'expected': fill([(1,3),(9,10),(12,14)])}]

############################### Implementation #################################
def generate(**kwargs):
    from gMiner.manipulate import fusion
    for f in fusion(meta_generate(**kwargs)): yield f

def meta_generate(n_tracks, fn, win_size):
    from gMiner.manipulate import fusion
    tracks = n_tracks
    sentinel = (sys.maxint, sys.maxint, 0.0)
    tracks = [fusion(t) for t in tracks]
    tracks = [common.sentinelize(t, sentinel) for t in tracks]

    N = len(tracks)
    init = [tracks[i].next() for i in range(N)]
    for i in xrange(N-1,-1,-1):
        if init[i] == sentinel: # empty track
            N-=1
            tracks.pop(i)
            init.pop(i)
    available_tracks = range(N-1,-1,-1)
    activity = [False]*N
    current = []
    current.extend([(init[i][0],i) for i in range(N)])
    current.extend([(init[i][1],i) for i in range(N)])
    current.sort()

    start = current[0][0]
    while current[0][0] == start:
        activity[current[0][1]] = True
        z = current.pop(0)

    k=1
    while available_tracks:
        # load *win_size* bp in memory
        to_remove = []
        for i in available_tracks:
            a = [0,0]
            limit = k*win_size
            while a[1] < limit:
                a = tracks[i].next()
                if a == sentinel:
                    to_remove.append(i)
                    break
                current.append((a[0],i))
                current.append((a[1],i))
        for i in to_remove:
            available_tracks.remove(i)
        current.sort()

        # calculate boolean values for start-next interval
        while current and current[0][0] < limit:
            next = current[0][0]
            res = fn(activity)
            if res:
                yield (start,next,'',0.0,0)
            while current and current[0][0] == next:
                i = current[0][1]
                activity[i] = not(activity[i])
                z = current.pop(0)
            start = next
        k+=1




"""
Contains URL download tests.
"""

# Internal modules #
import track

# Unittesting module #
try:
    import unittest2 as unittest
except ImportError:
    import unittest

# Nosetest flag #
__test__ = True

###################################################################################
class TestDownload(unittest.TestCase):
    def runTest(self):
        url = "http://salt.epfl.ch/BED/sinclair/genomic/ChIP.bedGraph"
        with track.load(url) as t:
            got = t.count('chrY')
            expected = 577
        self.assertEqual(got, expected)

#-----------------------------------#
# This code was written by the BBCF #
# http://bbcf.epfl.ch/              #
# webmaster.bbcf@epfl.ch            #
#-----------------------------------#



"""
This module implements the WIG serialization.
"""

# Internal modules #
from track.serialize import Serializer
from track.common import format_float

# Constants #
all_fields = ['start', 'end', 'score']

################################################################################
class SerializerWIG(Serializer):
    format = 'wig'

    def __enter__(self):
        self.file = open(self.path, 'w')
        self.indices = None
        self.previous_end = None
        self.previous_span = None
        return self

    def __exit__(self, errtype, value, traceback):
        self.file.close()

    def defineFields(self, fields):
        # Check that we have what we need #
        if not 'start' in fields: self.error("You tried to write a WIG file without a start field.")
        if not 'end' in fields:   self.error("You tried to write a WIG file without a end field.")
        if not 'score' in fields: self.error("You tried to write a WIG file without a score field.")
        # Check for the simple case  #
        if fields == tuple(all_fields): return
        # What indexes should we take #
        self.indices = [fields.index(f) for f in all_fields]

    def newTrack(self, info=None, name=None):
        if not info: info = {}
        info['type'] = 'wig'
        info['converted_by'] = __package__
        self.file.write("track " + ' '.join([k + '="' + v + '"' for k, v in info.items()]) + '\n')
        self.tracks.append(self.path)

    def newFeature(self, chrom, feature):
        # Put the fields in the right order #
        if self.indices: f = [feature[i] for i in self.indices]
        else:            f = list(feature)
        start, end, span, score = f[0], f[1], f[1]-f[0], format_float(f[2])
        # Look ahead #
        if not self.previous_end:
            self.writeFixedStep(chrom, start, end, span, score)
            return
        # If we have the same span just add the score #
        if start == self.previous_end and span == self.previous_span:
            self.file.write(score + '\n')
            self.previous_end += span
        else:
            self.writeFixedStep(chrom, start, end, span, score)

    def writeFixedStep(self, chrom, start, end, span, score):
        self.file.write("fixedStep chrom=%s start=%s span=%s\n%s\n" % (chrom, start, span, score))
        self.previous_end = end
        self.previous_span = span

#-----------------------------------#
# This code was written by the BBCF #
# http://bbcf.epfl.ch/              #
# webmaster.bbcf@epfl.ch            #
#-----------------------------------#



# coding: utf-8

################################### Modules ####################################
import sys
from track import common
from track.test import samples

################################# Description ##################################
label           = 'Overlap'
short_name      = 'Overlap'
long_name       = 'Pieces of overlap between two tracks (boolean AND)'
input_tracks    = [{'key':'X', 'position':1, 'kind':'single',
                    'fields':['start','end','name','score','strand','...']},
                   {'key':'Y', 'position':2, 'kind':'single',
                    'fields':['start','end','name','score','strand','...']}]
input_args      = []
input_meta      = [{'key':'l', 'position':2, 'type': int, 'kind':'chrom_len'}]
output_tracks   = [{'position':1, 'fields': ['start','end','name','score','strand','...']}]
tracks_collapse = None
fields_collapse = None
chroms_collapse = 'intersection'

################################ Documentation #################################
tooltip = \
"""
The ``overlap`` manipulation computes the overlap between two tracks
returning new features that exactly match the overlapping zones.
This is equivalent to the boolean `AND` operation.
"""

tickets = ['https://github.com/bbcf/gMiner/issues/4']

################################## Examples ####################################
numeric_example = \
"""
X [start,end]: (0,20)
Y [start,end]: (10,30)
R [start,end]: (10,20)
"""

visual_example = \
u"""
X: ───▤▤▤▤▤▤▤▤▤─────────────────▤▤▤▤▤▤▤──────
Y: ─────────▤▤▤▤▤▤▤────▤▤▤▤▤─────────────────
R: ─────────▤▤▤──────────────────────────────
"""

#################################### Tests #####################################
tests = [
    {'tracks':     {'X': samples['small_features'][2]['sql'],
                    'Y': samples['small_features'][3]['sql']},
     'chromosome': 'chrI',
     'expected':  [( 15,  20, u'NameA + Name1',  0.1,  0),
                   ( 32,  38, u'NameB + Name2',  0.2,  1),
                   ( 95, 100, u'NameD + Name5',  0.0,  1),
                   (110, 115, u'Name6 + NameD',  0.2,  1),
                   (130, 135, u'Name7 + NameE',  0.4,  0),
                   (140, 145, u'NameF + Name7',  0.4,  0),
                   (185, 190, u'NameG + Name8',  0.2,  1),
                   (185, 195, u'NameG + Name9',  0.2,  1),
                   (210, 215, u'Name10 + NameH', 0.45-5e-17, 1),
                   (215, 220, u'NameI + Name10', 0.25, 1),
                   (235, 240, u'NameJ + Name11', 0.1,  1),
                   (250, 254, u'Name12 + NameJ', 0.15+2e-17, 1),
                   (252, 258, u'NameK + Name12', 0.25, 1),
                   (256, 260, u'NameL + Name12', 0.2,  1),
                   (270, 275, u'Name13 + NameL', 0.1,  1)]}]

############################### Implementation #################################
def make_name(a, b):
    if a and b: return a + ' + ' + b
    elif a: return a
    return b

def make_feature(a, b):
    return (max(a[0],b[0]),
            min(a[1],b[1]),
            make_name(a[2], b[2]),
            (a[3]+b[3])/2.0,
            a[4]==b[4] and b[4] or 0) + b[5:]

def generate(X, Y, l):
    """Inspired from:
    fjoin: Simple and Efficient Computation of Feature Overlap
    Journal of Computational Biology, 13(8), Oct. 2006, pp 1457-1464
    """
    # Preparation #
    sentinel = (sys.maxint, sys.maxint)
    X = common.sentinelize(X, sentinel)
    Y = common.sentinelize(Y, sentinel)
    x = X.next()
    y = Y.next()
    Wx = []
    Wy = []
    # Core loop stops when both x and y are at the sentinel
    while x[0] != sentinel or y[0] != sentinel:
        # Take the leftmost current feature and scan it against the other window
        if x[0] < y[0]:
            # Remove features from the y window that are left of x
            Wy = [f for f in Wy if f[1] > x[0]]
            # Yield new features with all overlaps of x in Wy
            for f in [f for f in Wy if f[1] > x[0] and x[1] > f[0]]: yield make_feature(x, f)
            # Put x in the window only if it is not left of y
            if x[1] >= y[0]: Wx.append(x)
            # Advance current x feature
            x = X.next()
        else:
            # Remove features from the x window that are left of y
            Wx = [f for f in Wx if f[1] > y[0]]
            # Yield new features with all overlaps of y in Wx
            for f in [f for f in Wx if f[1] > y[0] and y[1] > f[0]]: yield make_feature(y, f)
            # Put y in the window only if it is not left of x
            if y[1] >= x[0]: Wy.append(y)
            # Advance current y feature
            y = Y.next()



"""
This module implements the SGA serialization.

http://ccg.vital-it.ch/chipseq/doc/chipseq_tech.php
"""

# Internal modules #
from track.serialize import Serializer

# Functions #
def int_to_strand(num):
    num = int(num)
    if num == 1: return  '+'
    if num == -1: return '-'
    return '0'

# Constants #
all_fields = ['start', 'end', 'name', 'strand', 'score']
defaults   = [None, None, 'signal', '0', '0']

################################################################################
class SerializerSGA(Serializer):
    format = 'sga'

    def __enter__(self):
        self.file = open(self.path, 'w')
        self.indices = []
        self.current_chrom = None
        return self

    def __exit__(self, errtype, value, traceback):
        self.file.close()

    def defineFields(self, fields):
        # Check for required fields #
        if 'start' not in fields or 'end' not in fields:
            self.error("You tried to write an SGA file without a start and end field")
        # Check that all fields are legal #
        for f in fields:
            if f not in all_fields:
                message = "You tried to write a '%s' field to a SGA file. Possible fields are: %s"
                self.error(message % (f, all_fields))
        # Find indicies #
        for i,f in enumerate(all_fields):
            if f in fields: self.indices.append(fields.index(f))
            else: self.indices.append(defaults[i])

    def newTrack(self, info=None, name=None):
        if not info: info = {}
        info['type'] = 'sga'
        info['converted_by'] = __package__
        self.tracks.append(self.path)

    def newFeature(self, chrom, feature):
        # The chromsoome name must be NC type #
        if chrom != self.current_chrom:
            name = self.assembly.guess_chromosome_name(chrom)
            self.current_chrom = self.chrmeta[name]['refseq']
        # Put the fields in the right order #
        feature = [feature[i] if isinstance(i,int) else i for i in self.indices]
        start, end, name, strand, score = feature
        # Convert the score #
        try: score = str(int(score))
        except ValueError: pass
        # Convert the strand #
        try: strand = int_to_strand(strand)
        except ValueError: pass
        # Write one line #
        for i in range(end-start):
            self.file.write('\t'.join((self.current_chrom, name, str(start+1+i), strand, score)) + '\n')

#-----------------------------------#
# This code was written by the BBCF #
# http://bbcf.epfl.ch/              #
# webmaster.bbcf@epfl.ch            #
#-----------------------------------#



"""
Contains tests for ``t.delete_fields()``.
"""

# Built-in modules #
import os, shutil

# Internal modules #
import track
from track.test import samples
from track.common import temporary_path

# Unittesting module #
try:
    import unittest2 as unittest
except ImportError:
    import unittest

# Nosetest flag #
__test__ = True

###################################################################################
class Test(unittest.TestCase):
    def runTest(self):
        orig_path = samples['small_features'][2]['sql']
        test_path = temporary_path('.sql')
        shutil.copy(orig_path, test_path)
        with track.load(test_path) as t:
            t.delete_fields(['name','strand'])
            got = list(t.read())
        expected = [('chrI',  10,  20, 0.1),
                    ('chrI',  30,  40, 0.2),
                    ('chrI',  50,  60, 0.1),
                    ('chrI',  70,  80, 0.2),
                    ('chrI',  90, 100, 0.0),
                    ('chrI', 110, 120, 0.4),
                    ('chrI', 130, 150, 0.4),
                    ('chrI', 180, 190, 0.1),
                    ('chrI', 180, 200, 0.1),
                    ('chrI', 210, 220, 0.2),
                    ('chrI', 230, 240, 0.1),
                    ('chrI', 250, 260, 0.2),
                    ('chrI', 270, 280, 0.0),
                    ('chrI', 290, 300, 0.7)]
        self.assertEqual(got, expected)
        # Clean up #
        os.remove(test_path)

#-----------------------------------#
# This code was written by the BBCF #
# http://bbcf.epfl.ch/              #
# webmaster.bbcf@epfl.ch            #
#-----------------------------------#



"""
This module implements the parsing of the track object. The specification can be found here:

http://bbcf.epfl.ch/twiki/bin/view/BBCF/SqLite

In this case the self.path attribute can be either
a Track object or the path to an SQL file.
"""

# Internal modules #
from track import Track, load
from track.parse import Parser

################################################################################
class ParserSQL(Parser):
    format = 'sql'
    def parse(self):
        # Core function #
        def read_whole_track(t):
            self.handler.defineFields(t.fields)
            self.handler.newTrack(t.info)
            if t.info.get('assembly'): self.handler.defineAssembly(t.info.get('assembly'))
            else: self.handler.defineChrmeta(t.chrmeta)
            for chrom in t:
                for feature in t.read(chrom):
                    self.handler.newFeature(chrom, feature)
        # Check param type #
        if isinstance(self.path, Track):
            read_whole_track(self.path)
        else:
            with load(self.path, 'sql', readonly=True) as t: read_whole_track(t)

#-----------------------------------#
# This code was written by the BBCF #
# http://bbcf.epfl.ch/              #
# webmaster.bbcf@epfl.ch            #
#-----------------------------------#
